import sys
import logging


def get_logger(logger_name='logger', fname=None, debug=False):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    level = logging.DEBUG if debug else logging.INFO

    # LOG format
    fmt = "%(asctime)-15s %(levelname)-5s %(message)s"
    datefmt = "[%a %Y-%m-%d %H:%M:%S]"
    formatter = logging.Formatter(fmt, datefmt)

    # LOG file
    if fname is not None:
        fh = open(fname, 'w')
        fh.close()
        file_handler = logging.FileHandler(fname)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    # LOG console
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setLevel(level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    return logger



import os
import sys
import click
import importlib.metadata
from datetime import datetime
from porechopx import porechop, adapters, misc
from porechopx.parser import FastqChoper, FastqReader
from porechopx.logger import get_logger
__version__ = importlib.metadata.version('porechopx')


@click.command(help='PorechopX: a tool for finding adapters in Oxford Nanopore reads, trimming '
                    'them from the ends and splitting reads with internal adapters',
               context_settings={'show_default': True})
@click.version_option(__version__)
# Main options
@click.option('-i', '--input', 'input_filepath', required=True,
              help='FASTA/FASTQ of input reads or a directory which will be '
                   'recursively searched for FASTQ files')
@click.option('-o', '--output', 'output_filepath',
              help='Filename for FASTA or FASTQ of trimmed reads (if not set, '
                   'trimmed reads will be printed to stdout)')
@click.option('--barcode_stats_csv', default=None,
              help='Path to a csv file with start/ end/ middle barcode names '
                   'and percentage identities for each given read ( if not set, '
                   'no information will be printed)')
@click.option('--format', 'out_format', type=click.Choice(['auto', 'fasta', 'fastq', 'fasta.gz', 'fastq.gz']),
              default='auto',
              help='Output format for the reads - if auto, the '
                   'format will be chosen based on the output filename or the input '
                   'read format')
@click.option('-v', '--verbosity', type=int, default=1,
              help='Level of progress information: 0 = none, 1 = some, 2 = lots, '
                   '3 = full - output will go to stdout if reads are saved to '
                   'a file and stderr if reads are printed to stdout')
@click.option('-t', '--threads', type=int, default=lambda: min(os.cpu_count(), 16),
              help='Number of threads to use for adapter alignment')
@click.option('-c', '--chunk_size', type=int, default=10_000,
              help='Number of reads per chunk')
# Barcode binning settings:
#   Control the binning of reads based on barcodes (i.e. barcode demultiplexing)
@click.option('-b', '--barcode_dir',
              help='Reads will be binned based on their barcode and saved to '
                   'separate files in this directory (incompatible with --output)')
@click.option('--barcode_labels', is_flag=True,
              help='Reads will have a label added to their header with their barcode')
@click.option('--extended_labels', is_flag=True,
              help='Reads will have an extended label added to their header with the '
                   'barcode_call (if any), the best start/ end barcode hit and their '
                   'identities, and whether a barcode is found in middle of read. '
                   '(Dependent on --barcode_labels).')
@click.option('--native_barcodes', is_flag=True,
              help='Only attempts to match the 24 native barcodes')
@click.option('--pcr_barcodes', is_flag=True,
              help='Only attempts to match the 96 PCR barcodes')
@click.option('--rapid_barcodes', is_flag=True,
              help='Only attempts to match the 12 rapid barcodes')
@click.option('--limit_barcodes_to', multiple=True,
              help='Specify a list of barcodes to look for (numbers refer to native, PCR or rapid)')
@click.option('--custom_barcodes',
              help='CSV file containing custom barcode sequences')
@click.option('--barcode_threshold', type=float, default=75.0,
              help='A read must have at least this percent identity to a barcode to be binned')
@click.option('--barcode_diff', type=float, default=5.0,
              help="If the difference between a read's best barcode identity and its second-best "
                   "barcode identity is less than this value, it will not be put in a barcode bin "
                   "(to exclude cases which are too close to call)")
@click.option('--require_two_barcodes', is_flag=True,
              help='Reads will only be put in barcode bins if they have a strong match for the '
                   'barcode on both their start and end (default: a read can be binned with '
                   'a match at its start or end)')
@click.option('--untrimmed', is_flag=True,
              help='Bin reads but do not trim them (default: trim the reads)')
@click.option('--discard_unassigned', is_flag=True,
              help='Discard unassigned reads (instead of creating a "none" bin)')
# Adapter search settings:
#   Control how the program determines which adapter sets are present
@click.option('--adapter_threshold', type=float, default=90.0,
              help='An adapter set has to have at least this percent identity to be labelled as '
                   'present and trimmed off (0 to 100)')
@click.option('--check_reads', 'check_read_count', type=int, default=10000,
              help='This many reads will be aligned to all possible adapters to determine which '
                   'adapter sets are present')
@click.option('--scoring_scheme', type=str, default='3,-6,5,2',
              help='Comma-delimited string of alignment scores: '
                   'match, mismatch, gap open, gap extend')
# End adapter settings:
#   Control the trimming of adapters from read ends
@click.option('--end_size', type=int, default=150,
              help='The number of base pairs at each end of the read which will be searched for '
                   'adapter sequences')
@click.option('--min_trim_size', type=int, default=10,
              help='Adapter alignments smaller than this will be ignored')
@click.option('--extra_end_trim', type=int, default=2,
              help='This many additional bases will be removed next to adapters found at '
                   'the ends of reads')
@click.option('--end_threshold', type=float, default=75.0,
              help='Adapters at the ends of reads must have at least this percent identity '
                   'to be removed (0 to 100)')
# Middle adapter settings:
#   Control the splitting of read from middle adapters
@click.option('--no_split', is_flag=True,
              help='Skip splitting reads based on middle adapters (default: split reads when '
                   'an adapter is found in the middle)')
@click.option('--discard_middle', is_flag=True,
              help='Reads with middle adapters will be discarded (default: reads with '
                   'middle adapters are split)')
@click.option('--middle_threshold', type=float, default=90.0,
              help='Adapters in the middle of reads must have at least this percent identity '
                   'to be found (0 to 100)')
@click.option('--extra_middle_trim_good_side', type=int, default=10,
              help='This many additional bases will be removed next to middle adapters '
                   'on their "good" side')
@click.option('--extra_middle_trim_bad_side', type=int, default=100,
              help='This many additional bases will be removed next to middle adapters '
                   'on their "bad" side')
@click.option('--min_split_read_size', type=int, default=1000,
              help='Post-split read pieces smaller than this many base pairs '
                   'will not be outputted')
def main(input_filepath, output_filepath, barcode_stats_csv, out_format, verbosity, threads,
         chunk_size, barcode_dir, barcode_labels, extended_labels,
         native_barcodes, pcr_barcodes, rapid_barcodes, limit_barcodes_to, custom_barcodes,
         barcode_threshold, barcode_diff, require_two_barcodes, untrimmed, discard_unassigned,
         adapter_threshold, check_read_count, scoring_scheme,
         end_size, min_trim_size, extra_end_trim, end_threshold, no_split,
         discard_middle, middle_threshold, extra_middle_trim_good_side, extra_middle_trim_bad_side,
         min_split_read_size):

     # Checking arguments
     try:
          scoring_scheme_vals = [int(x) for x in scoring_scheme.split(',')]
     except ValueError:
          sys.exit('Error: incorrectly formatted scoring scheme')
     if len(scoring_scheme_vals) != 4:
          sys.exit('Error: incorrectly formatted scoring scheme')

     if barcode_dir and output_filepath:
          sys.exit('Error: only one of the following options may be used: --output, --barcode_dir')

     if untrimmed and not barcode_dir:
          sys.exit('Error: --untrimmed can only be used with --barcode_dir')

     if barcode_dir:
          discard_middle = True

     if threads < 1:
          sys.exit('Error: at least one thread required')

     # Program starts
     logger = get_logger("porechopx", debug=verbosity > 1)

     start_time = datetime.now()

     search_adapters = None
     matching_sets = None
     forward_or_reverse_barcodes = None

     fq = FastqReader(input_filepath, check_read_count, False)
     check_reads = fq.__next__()
     fq.close()

     # Check adapters
     if native_barcodes or pcr_barcodes or rapid_barcodes:
          if native_barcodes:
               logger.info('Using native barcodes')
               # construct a smaller set of search adapters with only the 24 barcodes to speed up
               # the initial step
               barcodes_set = adapters.NATIVE_BARCODES
               forward_or_reverse_barcodes = 'reverse'
               # barcode_adapters = []
               # for i in range(1, 25):
               #     barcode_adapters.append(make_full_native_barcode_adapter(i))
          elif pcr_barcodes:
               logger.info('Using PCR barcodes')
               barcodes_set = adapters.PCR_BARCODES
               forward_or_reverse_barcodes = 'forward'
          else:
               logger.info('Using rapid barcodes')
               barcodes_set = adapters.RAPID_BARCODES
               forward_or_reverse_barcodes = 'forward'

          if limit_barcodes_to:
               logger.info(f'Limiting barcodes to : {limit_barcodes_to}')

          logger.info(f'{len(barcodes_set)} barcodes in search set')

          matching_sets = []
          if limit_barcodes_to:
               for barcode_number in limit_barcodes_to:
                    if barcode_number < 1 or barcode_number > len(barcodes_set):
                         logger.error('Barcode number out of range of chosen set (1-24 for native, '
                                      '1-12 for rapid, 1-96 for PCR)')
                         sys.exit(1)
                    matching_sets.append(barcodes_set[barcode_number - 1])
          else:
               matching_sets.extend(barcodes_set)

          check_barcodes = True

     else:
          if limit_barcodes_to:
               logger.error('To limit search to specific barcodes, specify whether using native, '
                            'PCR, or rapid barcodes')
               sys.exit(1)

          if custom_barcodes:
               matching_sets = adapters.load_custom_barcodes(custom_barcodes)
               logger.info('Using custom barcodes')
               logger.info(f'{len(matching_sets)} barcodes in search set')

          if not matching_sets:
               if not search_adapters:
                    # just add all of the adapters
                    search_adapters = adapters.NATIVE_BARCODES
                    search_adapters.extend(adapters.PCR_BARCODES)
                    search_adapters.extend(adapters.RAPID_BARCODES)
                    search_adapters.extend(adapters.OTHER_ADAPTERS)

               matching_sets = porechop.find_matching_adapter_sets(
                    check_reads, verbosity, end_size,
                    scoring_scheme_vals, adapter_threshold,
                    threads, search_adapters
               )
               matching_sets = porechop.exclude_end_adapters_for_rapid(matching_sets)
               matching_sets = porechop.fix_up_1d2_sets(matching_sets)
               matching_sets = porechop.add_full_barcode_adapter_sets(matching_sets)
               porechop.display_adapter_set_results(search_adapters, matching_sets)

               if barcode_dir or barcode_labels:
                    forward_or_reverse_barcodes = porechop.choose_barcoding_kit(matching_sets)

          if matching_sets:
               check_barcodes = (barcode_dir is not None or barcode_labels is not False)

     # Perform adapter & barcode trimming
     if matching_sets:
          if no_split:
               logger.info('Trimming adapters from read ends\n')
          else:
               verb = 'discarding' if discard_middle else 'splitting'
               logger.info(f'Trimming adapters from read ends, and {verb} reads containing middle adapters\n')

          name_len = max(max(len(x.start_sequence[0]) for x in matching_sets),
                         max(len(x.end_sequence[0]) if x.end_sequence else 0 for x in matching_sets))
          for matching_set in matching_sets:
               sys.stderr.write('  ' + matching_set.start_sequence[0].rjust(name_len) + ': ' +
                    misc.red(matching_set.start_sequence[1]) + '\n')
               if matching_set.end_sequence:
                    sys.stderr.write('  ' + matching_set.end_sequence[0].rjust(name_len) + ': ' +
                         misc.red(matching_set.end_sequence[1]) + '\n')
          sys.stderr.write('\nthreads: ' + str(threads) + '\n\n')

          # Multiprocess trimming
          FastqChoper(
               input_filepath, matching_sets, verbosity, end_size,
               extra_end_trim, end_threshold,
               scoring_scheme_vals, min_trim_size,
               threads, check_barcodes, barcode_threshold,
               barcode_diff, require_two_barcodes,
               forward_or_reverse_barcodes, middle_threshold,
               extra_middle_trim_good_side, extra_middle_trim_bad_side,
               discard_middle, discard_unassigned, no_split,
               out_format, output_filepath, barcode_stats_csv, min_split_read_size,
               barcode_dir, barcode_labels, extended_labels, untrimmed,
               chunk_size, 10,
          )

     else:
          logger.info('No adapters found - output reads are unchanged from input reads\n')

     # Finish
     time = datetime.now() - start_time
     logger.info(f'Time taken:  {time}')


if __name__ == '__main__':
    main()



"""
Copyright 2017 Ryan Wick (rrwick@gmail.com)
https://github.com/rrwick/Porechop

This module contains miscellaneous functions used by Porechop.

This file is part of Porechop. Porechop is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Porechop is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Porechop. If
not, see <http://www.gnu.org/licenses/>.
"""

import sys
import os
import gzip
import re
import textwrap
import shutil
import argparse


def float_to_str(num, decimals, max_num=0):
    """
    Converts a number to a string. Will add left padding based on the max value to ensure numbers
    align well.
    """
    if decimals == 0:
        return int_to_str(int(round(num)), max_num=max_num)
    if num is None:
        num_str = 'n/a'
    else:
        num_str = '%.' + str(decimals) + 'f'
        num_str = num_str % num
        parts = num_str.split('.')
        before_decimal = parts[0]
        after_decimal = parts[1]
        num_str = int_to_str(int(before_decimal)) + '.' + after_decimal
    if max_num > 0:
        max_str = float_to_str(max_num, decimals)
        num_str = num_str.rjust(len(max_str))
    return num_str


def int_to_str(num, max_num=0):
    """
    Converts a number to a string. Will add left padding based on the max value to ensure numbers
    align well.
    """
    if num is None:
        num_str = 'n/a'
    else:
        num_str = '{:,}'.format(num)
    max_str = '{:,}'.format(int(max_num))
    return num_str.rjust(len(max_str))


def get_compression_type(filename):
    """
    Attempts to guess the compression (if any) on a file using the first few bytes.
    http://stackoverflow.com/questions/13044562
    """
    magic_dict = {'gz': (b'\x1f', b'\x8b', b'\x08'),
                  'bz2': (b'\x42', b'\x5a', b'\x68'),
                  'zip': (b'\x50', b'\x4b', b'\x03', b'\x04')}
    max_len = max(len(x) for x in magic_dict)

    unknown_file = open(filename, 'rb')
    file_start = unknown_file.read(max_len)
    unknown_file.close()
    compression_type = 'plain'
    for filetype, magic_bytes in magic_dict.items():
        if file_start.startswith(magic_bytes):
            compression_type = filetype
    if compression_type == 'bz2':
        sys.exit('Error: cannot use bzip2 format - use gzip instead')
    if compression_type == 'zip':
        sys.exit('Error: cannot use zip format - use gzip instead')
    return compression_type


def get_sequence_file_type(filename):
    """
    Determines whether a file is FASTA or FASTQ.
    """
    if not os.path.isfile(filename):
        sys.exit('Error: could not find ' + filename)
    if get_compression_type(filename) == 'gz':
        open_func = gzip.open
    else:  # plain text
        open_func = open

    with open_func(filename, 'rt') as seq_file:
        try:
            first_char = seq_file.read(1)
        except UnicodeDecodeError:
            first_char = ''

    if first_char == '>':
        return 'FASTA'
    elif first_char == '@':
        return 'FASTQ'
    else:
        raise ValueError('File is neither FASTA or FASTQ')


def load_fasta_or_fastq(filename):
    """
    Returns a list of tuples (header, seq) for each record in the fasta/fastq file.
    """
    try:
        file_type = get_sequence_file_type(filename)
        if file_type == 'FASTA':
            return load_fasta(filename), 'FASTA'
        else:  # FASTQ
            return load_fastq(filename), 'FASTQ'
    except IndexError:
        sys.exit('\nError: ' + filename + ' could not be parsed - is it formatted correctly?')


def load_fasta(fasta_filename):
    """
    Returns a list of tuples (header, seq) for each record in the fasta file.
    """
    if get_compression_type(fasta_filename) == 'gz':
        open_func = gzip.open
    else:  # plain text
        open_func = open
    fasta_seqs = []
    with open_func(fasta_filename, 'rt') as fasta_file:
        name = ''
        sequence = ''
        for line in fasta_file:
            line = line.strip()
            if not line:
                continue
            if line[0] == '>':  # Header line = start of new contig
                if name:
                    fasta_seqs.append((name.split()[0], sequence, name))
                    sequence = ''
                name = line[1:]
            else:
                sequence += line
        if name:
            fasta_seqs.append((name.split()[0], sequence, name))
    return fasta_seqs


def load_fastq(fastq_filename):
    """
    Returns a list of tuples (header, seq) for each record in the fastq file.
    """
    if get_compression_type(fastq_filename) == 'gz':
        open_func = gzip.open
    else:  # plain text
        open_func = open
    reads = []
    with open_func(fastq_filename, 'rt') as fastq:
        for line in fastq:
            full_name = line.strip()[1:]
            short_name = full_name.split()[0]
            sequence = next(fastq).strip()
            spacer = next(fastq).strip()
            qualities = next(fastq).strip()
            reads.append((short_name, sequence, spacer, qualities, full_name))
    return reads


def print_table(table, print_dest, alignments='', max_col_width=30, col_separation=3, indent=2,
                row_colour=None, sub_colour=None, row_extra_text=None, leading_newline=False,
                subsequent_indent='', return_str=False, header_format='underline',
                hide_header=False, fixed_col_widths=None, left_align_header=True,
                bottom_align_header=True):
    """
    Args:
        table: a list of lists of strings (one row is one list, all rows should be the same length)
        print_dest: either sys.stdout or sys.stderr - where the table will be printed
        alignments: a string of L and R, indicating the alignment for each row
        max_col_width: values longer than this will be wrapped
        col_separation: the number of spaces between columns
        indent: the number of spaces between the table and the left side of the terminal
        row_colour: a dictionary of row indices and their colour names
        sub_colour: a dictionary of values to colour names for which the text colour will be set
        row_extra_text: a dictionary of row indices and extra text to display after the row
        leading_newline: if True, the function will print a blank line above the table
        subsequent_indent: this string will be added to the start of wrapped text lines
        return_str: if True, this function will return a string of the table instead of printing it
        header_format: the formatting (colour, underline, etc) of the header line
        hide_header: if True, the header is not printed
        fixed_col_widths: a list to specify exact column widths (automatic if not used)
        left_align_header: if False, the header will follow the column alignments
        bottom_align_header: if False, the header will align to the top, like other rows
    """
    column_count = len(table[0])
    table = [x[:column_count] for x in table]
    table = [x + [''] * (column_count - len(x)) for x in table]
    if row_colour is None:
        row_colour = {}
    if sub_colour is None:
        sub_colour = {}
    if row_extra_text is None:
        row_extra_text = {}
    if leading_newline:
        print('', file=print_dest)

    # Ensure the alignments string is the same length as the column count
    alignments += 'L' * (column_count - len(alignments))
    alignments = alignments[:column_count]

    if fixed_col_widths is not None:
        col_widths = fixed_col_widths
    else:
        col_widths = [0] * column_count
        for row in table:
            col_widths = [min(max(col_widths[i], len_without_format(x)), max_col_width)
                          for i, x in enumerate(row)]
    separator = ' ' * col_separation
    indenter = ' ' * indent
    full_table_str = ''
    for i, row in enumerate(table):
        row = [str(x) for x in row]
        if hide_header and i == 0:
            continue

        if fixed_col_widths is not None:
            wrapped_row = []
            for col, fixed_width in zip(row, fixed_col_widths):
                wrapper = textwrap.TextWrapper(subsequent_indent=subsequent_indent,
                                               width=fixed_width)
                wrapped_row.append(wrapper.wrap(col))
        else:
            wrapper = textwrap.TextWrapper(subsequent_indent=subsequent_indent, width=max_col_width)
            wrapped_row = [wrapper.wrap(x) for x in row]

        row_rows = max(len(x) for x in wrapped_row)
        if i == 0 and bottom_align_header:
            wrapped_row = [[''] * (row_rows - len(x)) + x for x in wrapped_row]

        for j in range(row_rows):
            row_line = [x[j] if j < len(x) else '' for x in wrapped_row]
            aligned_row = []
            for value, col_width, alignment in zip(row_line, col_widths, alignments):
                if alignment == 'L' or (i == 0 and left_align_header):
                    aligned_row.append(value.ljust(col_width))
                elif alignment == 'C':
                    aligned_row.append(value.center(col_width))
                else:
                    aligned_row.append(value.rjust(col_width))
            row_str = separator.join(aligned_row)
            if i in row_extra_text:
                row_str += row_extra_text[i]
            if i == 0 and header_format:
                row_str = colour(row_str, header_format)
            if i in row_colour:
                row_str = colour(row_str, row_colour[i])
            for text, colour_name in sub_colour.items():
                row_str = row_str.replace(text, colour(text, colour_name))
            if j < row_rows - 1 and UNDERLINE in row_str:
                row_str = re.sub('\033\[4m', '', row_str)
            if return_str:
                full_table_str += indenter + row_str + '\n'
            else:
                print(indenter + row_str, flush=True, file=print_dest)
    if return_str:
        return full_table_str


END_FORMATTING = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'
RED = '\033[31m'
GREEN = '\033[32m'
YELLOW = '\033[93m'
DIM = '\033[2m'


def colour(text, text_colour):
    bold_text = 'bold' in text_colour
    text_colour = text_colour.replace('bold', '')
    underline_text = 'underline' in text_colour
    text_colour = text_colour.replace('underline', '')
    text_colour = text_colour.replace('_', '')
    text_colour = text_colour.replace(' ', '')
    text_colour = text_colour.lower()
    if 'red' in text_colour:
        coloured_text = RED
    elif 'green' in text_colour:
        coloured_text = GREEN
    elif 'yellow' in text_colour:
        coloured_text = YELLOW
    elif 'dim' in text_colour:
        coloured_text = DIM
    else:
        coloured_text = ''
    if bold_text:
        coloured_text += BOLD
    if underline_text:
        coloured_text += UNDERLINE
    if not coloured_text:
        return text
    coloured_text += text + END_FORMATTING
    return coloured_text


def red(text):
    return RED + text + END_FORMATTING


def yellow(text):
    return YELLOW + text + END_FORMATTING

def bold(text):
    return BOLD + text + END_FORMATTING

def bold_underline(text):
    return BOLD + UNDERLINE + text + END_FORMATTING


def len_without_format(text):
    return len(remove_formatting(text))


def remove_formatting(text):
    return re.sub('\033.*?m', '', text)


def add_line_breaks_to_sequence(sequence, line_length):
    """
    Wraps sequences to the defined length.  All resulting sequences end in a line break.
    """
    if not sequence:
        return '\n'
    seq_with_breaks = ''
    pos = 0
    while pos < len(sequence):
        seq_with_breaks += sequence[pos:pos+line_length] + '\n'
        pos += line_length
    return seq_with_breaks


class MyHelpFormatter(argparse.HelpFormatter):
    """
    This is a custom formatter class for argparse. It allows for some custom formatting,
    in particular for the help texts with multiple options (like bridging mode and verbosity level).
    http://stackoverflow.com/questions/3853722
    """
    def __init__(self, prog):
        terminal_width = shutil.get_terminal_size().columns
        os.environ['COLUMNS'] = str(terminal_width)
        max_help_position = min(max(24, terminal_width // 3), 40)
        super().__init__(prog, max_help_position=max_help_position)

    def _get_help_string(self, action):
        help_text = action.help
        if action.default != argparse.SUPPRESS and 'default' not in help_text.lower() and \
                action.default is not None:
            help_text += ' (default: ' + str(action.default) + ')'
        return help_text


import os
import sys
import gzip
import logging
import subprocess
import multiprocessing
import numpy as np
from tqdm import tqdm
from pathlib import Path
from itertools import cycle
from typing import NamedTuple
from traceback import format_exc
from collections import defaultdict
from porechopx.porechop import NanoporeRead
from porechopx.misc import int_to_str, print_table
logger = logging.getLogger("porechopx")

# Nanopore read data class
Read = NamedTuple("Read", [('id', str), ('seq', str), ('qual', str)])


class FastqReaderPy(object):
    """
    A class to iterate over records in a FASTQ file.

    Each record is returned as a NamedTuple: (id, seq, qual)
    """
    def __init__(self, filepath, chunk_size=1_000_000, verbose=True) -> None:
        """
        Initialize the FastqReader with the path to the FASTQ file.
        """
        self.input = filepath
        self.chunk_size = chunk_size

        # Open handles
        self.file = open(self.input, 'rb')
        self.fh = gzip.open(self.file, 'rt')

        # Progress bar
        self.cursor = 0
        if verbose:
            self.pbar = tqdm(
                total=os.path.getsize(self.input),
                ncols=75, unit='b', unit_scale=True, unit_divisor=1024
            )
        else:
            self.pbar = None

    def __iter__(self):
        """
        Returns the iterator object itself.
        """
        return self

    def __next__(self):
        """
        Return the next record from the FASTQ file.
        """
        chunk = []
        for line in self.fh:
            # Load next read
            header = line.rstrip().lstrip("@")
            sequence = self.fh.readline().rstrip()
            separator = self.fh.readline().rstrip()
            quality = self.fh.readline().rstrip()

            # Push into chunk
            chunk.append(Read(header, sequence, quality))
            if len(chunk) >= self.chunk_size:
                break

        # Update progress
        processed = self.file.tell() - self.cursor
        if self.pbar is not None:
            self.pbar.update(processed)
        self.cursor += processed

        # End of file
        if not chunk:
            self.close()
            raise StopIteration

        return chunk

    def close(self):
        """
        Close the FASTQ file if it's still open.
        """
        if not self.fh.closed:
            self.fh.close()

        if not self.file.closed:
            self.file.close()

        if self.pbar is not None:
            self.pbar.update(self.pbar.total - self.pbar.n)
            self.pbar.close()
            sys.stderr.flush()


class FastqReaderRS(object):
    """
    A class to iterate over records in a FASTQ file. but using needletail

    Each record is returned as an needtail instance
        object (_type_): _description_
    """
    def __init__(self, filepath, chunk_size=1_000_000, verbose=True) -> None:
        """
        Initialize the FastqReaderX with the path to the FASTQ file.
        """
        from needletail import parse_fastx_file
        self.input = filepath
        self.chunk_size = chunk_size
        self.file = parse_fastx_file(filepath)

        # Progressbar
        if verbose:
            self.pbar = tqdm(
                total=os.path.getsize(self.input),
                ncols=75, unit='b', unit_scale=True, unit_divisor=1024
            )
        else:
            self.pbar = None

    def __iter__(self):
        """
        Returns the iterator object itself.
        """
        return self

    def __next__(self):
        """
        Return the next record from the FASTQ file.
        """
        try:
            chunk, processed = [], 0
            for record in self.file:
                chunk.append(Read(record.id, record.seq, record.qual))
                processed += len(record.id) + len(record.seq)*2 + 6 # Byte to store a read
                if len(chunk) >= self.chunk_size:
                    break

            # Update progress
            # A rough estimation for guppy basecalled reads = bytes / 2
            if self.pbar is not None:
                self.pbar.update(min(processed/2, self.pbar.total - self.pbar.n))

            # End of file
            if not chunk:
                self.close()
                raise StopIteration

        except KeyboardInterrupt:
            # Catch KeyboardInterrupt for terminating needletail iterator
            # https://stackoverflow.com/questions/21120947/catching-keyboardinterrupt-in-python-during-program-shutdown
            logger.error('Interrupted')
            try:
                sys.exit(130)
            except SystemExit:
                os._exit(130)

        return chunk

    def close(self):
        """
        Close the FASTQ file if it's still open.
        """
        if self.pbar is not None:
            self.pbar.update(self.pbar.total - self.pbar.n)
            self.pbar.close()
            sys.stderr.flush()


# Only use FastqReaderRS when needletail is corrected installed
try:
    import needletail
    FastqReader = FastqReaderRS
except ImportError:
    FastqReader = FastqReaderPy


class FastqChoper(object):
    """
    Class for trmming nanopore reads
    """
    def __init__(self, filepath, matching_sets, verbosity, end_size, extra_trim_size,
                 end_threshold, scoring_scheme_vals, min_trim_size,
                 threads, check_barcodes, barcode_threshold, barcode_diff,
                 require_two_barcodes, forward_or_reverse_barcodes,
                 middle_threshold, extra_middle_trim_good_side, extra_middle_trim_bad_side,
                 discard_middle, discard_unassigned, no_split,
                 out_format, output, barcode_stats_csv,
                 min_split_read_size,
                 barcode_dir, barcode_labels, extended_labels, untrimmed,
                 chunk_size=100_000, cache_size=10) -> None:

        # Global parameters
        self.input = filepath
        self.matching_sets = matching_sets
        self.verbosity = verbosity

        # Parameters for end trimming
        self.end_size = end_size
        self.extra_trim_size = extra_trim_size
        self.end_threshold = end_threshold
        self.scoring_scheme_vals = scoring_scheme_vals
        self.min_trim_size = min_trim_size
        self.threads = threads
        self.check_barcodes = check_barcodes
        self.barcode_threshold = barcode_threshold
        self.barcode_diff = barcode_diff
        self.require_two_barcodes = require_two_barcodes
        self.forward_or_reverse_barcodes = forward_or_reverse_barcodes

        # Parameters for middle splitting
        self.middle_threshold = middle_threshold
        self.extra_middle_trim_good_side = extra_middle_trim_good_side
        self.extra_middle_trim_bad_side = extra_middle_trim_bad_side
        self.discard_middle = discard_middle
        self.discard_unassigned = discard_unassigned
        self.no_split = no_split

        # Parameters for output
        self.out_format = out_format
        self.output = Path(output) if output is not None else None
        self.barcode_stats_csv = barcode_stats_csv
        self.min_split_size = min_split_read_size
        self.barcode_dir = Path(barcode_dir) if barcode_dir is not None else None
        self.barcode_labels = barcode_labels
        self.extended_labels = extended_labels
        self.untrimmed = untrimmed

        self.chunk_size = chunk_size

        # Saving output files
        trimmed_or_untrimmed = 'untrimmed' if self.untrimmed else 'trimmed'
        if self.barcode_dir is not None:
            verb = 'Saving'
            destination = 'barcode-specific files'
        elif self.output is None:
            verb = 'Outputting'
            destination = 'stdout'
        else:
            verb = 'Saving'
            destination = 'file'
        logger.info(f'{verb} {trimmed_or_untrimmed} reads to {destination}')

        if subprocess.getstatusoutput('which pigz')[0] == 0:
            logger.info('pigz found - using it to compress instead of gzip')
        else:
            logger.info('pigz not found - using gzip to compress')

        # Init multiprocessing
        manager = multiprocessing.Manager()
        self.exit = multiprocessing.Event()
        self.errors = manager.list()
        self.pool = []

        # Init queues for dandling input and output data
        self.queues_in = [multiprocessing.Queue(maxsize=cache_size) for _ in range(self.threads)]
        self.queues_out = [multiprocessing.Queue(maxsize=cache_size) for _ in range(self.threads)]

        # Init Workers
        self.pool = []
        for q_in, q_out in zip(self.queues_in, self.queues_out):
            p = multiprocessing.Process(target=self.worker, args=(q_in, q_out))
            self.pool.append(p)
            p.start()

        # Start loading fastq
        self.process_read = multiprocessing.Process(target=self.loader)
        self.process_read.start()
        self.process_write = multiprocessing.Process(target=self.writer)
        self.process_write.start()

        # Wait
        self.join()

    def loader(self):
        """
        Process for reading fastq into chunk, and pass into queues_in
        """
        try:
            # Iterator
            # TODO: infer data format
            fq = FastqReader(self.input, self.chunk_size)

            # Pass data into queues
            for chunk, q in zip(fq, cycle(self.queues_in)):
                # Exit if error
                if self.exit.is_set():
                    os._exit(1)

                q.put(chunk)

            # End of file
            for q in self.queues_in:
                q.put(None)

        except Exception as e:
            for q in self.queues_in:
                q.put(None)

            # Exit process if exception is caught
            pid = os.getpid()
            self.errors.append((pid, format_exc()))
            self.exit.set()
            os._exit(1)

    def worker(self, queue_in, queue_out):
        try:
            adapters = []
            start_sequence_names = set()
            end_sequence_names = set()
            for matching_set in self.matching_sets:
                adapters.append(matching_set.start_sequence)
                start_sequence_names.add(matching_set.start_sequence[0])
                if matching_set.end_sequence:
                    end_sequence_names.add(matching_set.end_sequence[0])
                    if matching_set.end_sequence[1] == matching_set.start_sequence[1]:
                        continue
                    adapters.append(matching_set.end_sequence)

            while True:
                # Exit if error
                if self.exit.is_set():
                    os._exit(1)

                # Process data
                chunk = queue_in.get()
                if chunk is None:
                    break

                results = []
                for _read in chunk:
                    read = NanoporeRead(_read.id, _read.seq, _read.qual)

                    # find_adapters_at_read_ends
                    read.find_start_trim(
                        self.matching_sets, self.end_size, self.extra_trim_size, self.end_threshold,
                        self.scoring_scheme_vals, self.min_trim_size, self.check_barcodes,
                        self.forward_or_reverse_barcodes)
                    read.find_end_trim(
                        self.matching_sets, self.end_size, self.extra_trim_size, self.end_threshold,
                        self.scoring_scheme_vals, self.min_trim_size, self.check_barcodes,
                        self.forward_or_reverse_barcodes)

                    if self.check_barcodes:
                        read.determine_barcode(
                            self.barcode_threshold, self.barcode_diff, self.require_two_barcodes)

                    # elif verbosity == 2:
                    #     sys.stderr.write(read.formatted_start_and_end_seq(end_size, extra_trim_size, check_barcodes),
                    #         file=print_dest)
                    # elif verbosity > 2:
                    #     print(read.full_start_end_output(end_size, extra_trim_size, check_barcodes),
                    #         file=print_dest)

                    # find_adapters_in_read_middles
                    if not self.no_split:
                        if self.discard_unassigned and read.barcode_call == "none":
                            # if we are discarding unassigned barcodes then there is no point in checking for middle adapters
                            continue

                        read.find_middle_adapters(
                            adapters, self.middle_threshold, self.extra_middle_trim_good_side,
                            self.extra_middle_trim_bad_side, self.scoring_scheme_vals,
                            start_sequence_names, end_sequence_names)

                        # if read.middle_adapter_positions and verbosity > 1:
                            # print(read.middle_adapter_results(
                                # verbosity), file=print_dest, flush=True)

                    results.append(read)

                queue_out.put(results)

            # Finish
            queue_out.put(None)

        except Exception as e:
            # Exit process if exception is caught
            queue_out.put(None)
            pid = os.getpid()
            self.errors.append((pid, format_exc()))
            self.exit.set()
            os._exit(1)

    def writer(self):
        try:
            # Barcode summary
            if self.barcode_stats_csv is not None:
                custom_output = open(self.barcode_stats_csv, "w")
                custom_output.write("name,start_time,barcode_call,start_name,start_id,end_name,end_id,middle_name,middle_id\n")

            # TODO: auto set output format
            self.out_format = 'fastq.gz'
            # if out_format == 'auto':
            #     if output is None:
            #         out_format = read_type.lower()
            #         if barcode_dir is not None and input_filename.lower().endswith('.gz'):
            #             out_format += '.gz'
            #     elif '.fasta.gz' in output.lower():
            #         out_format = 'fasta.gz'
            #     elif '.fastq.gz' in output.lower():
            #         out_format = 'fastq.gz'
            #     elif '.fasta' in output.lower():
            #         out_format = 'fasta'
            #     elif '.fastq' in output.lower():
            #         out_format = 'fastq'
            #     else:
            #         out_format = read_type.lower()

            gzipped_out = False
            gzip_command = 'gzip'
            if self.out_format.endswith('.gz') and \
                (self.barcode_dir is not None or self.output is not None):
                gzipped_out = True
                self.out_format = self.out_format[:-3]
                if subprocess.getstatusoutput('which pigz')[0] == 0:
                    gzip_command = f'pigz -p {self.threads}'

            # Output reads to barcode bins
            if self.barcode_dir is not None:
                if not self.barcode_dir.is_dir():
                    self.barcode_dir.mkdir(parents=True, exist_ok=True)
                barcode_files = {}
                barcode_read_counts, barcode_base_counts = defaultdict(int), defaultdict(int)

            # Output all reads to stdout
            elif self.output is None:
                pass

            # Output all reads to file
            else:
                if gzipped_out:
                    out_filename = self.output.with_name(f'TEMP_{os.getpid()}.{self.output.stem}')
                else:
                    out_filename = self.output
                out_file = open(out_filename, 'wt')

            # Iter through results
            total_reads = 0
            start_trim_total, start_trim_count = 0, 0
            end_trim_total, end_trim_count = 0, 0
            middle_trim_count = 0

            is_finish = False
            while True:
                # Exit if exception
                if self.exit.is_set():
                    os._exit(1)

                # Exit if finish
                if is_finish:
                    break

                # Get output
                for q in cycle(self.queues_out):
                    chunk = q.get()

                    # If finish
                    if chunk is None:
                        is_finish = True
                        break

                    # Summarize trimmed bases
                    total_reads += len(chunk)

                    start_trim_amount = np.array([read.start_trim_amount for read in chunk])
                    start_trim_total += start_trim_amount.sum()
                    start_trim_count += (start_trim_amount > 0).sum()

                    end_trim_amount = np.array([read.end_trim_amount for read in chunk])
                    end_trim_total += end_trim_amount.sum()
                    end_trim_count += (end_trim_amount > 0).sum()

                    middle_trim_count += sum([1 if read.middle_adapter_positions else 0 for read in chunk])

                    # Save to files
                    for read in chunk:
                        # Barcode information
                        if self.barcode_stats_csv is not None:
                            read.write_to_stats_csv(custom_output)

                        # Drop unassigned reads if specified
                        if self.discard_unassigned and read.barcode_call == 'none':
                            continue

                        # Prepare output fasta/fastq str
                        if self.out_format == 'fasta':
                            read_str = read.get_fasta(
                                self.min_split_size, self.discard_middle, self.untrimmed,
                                self.barcode_labels, self.extended_labels)
                        else:
                            read_str = read.get_fastq(
                                self.min_split_size, self.discard_middle, self.untrimmed,
                                self.barcode_labels, self.extended_labels)
                        if not read_str:
                            continue

                        # Output reads to barcode bin
                        if self.barcode_dir is not None:
                            barcode_name = read.barcode_call
                            if barcode_name not in barcode_files:
                                barcode_files[barcode_name] = open(
                                    self.barcode_dir / f'{barcode_name}.{self.out_format}', 'wt'
                                )
                            barcode_files[barcode_name].write(read_str)
                            barcode_read_counts[barcode_name] += 1
                            if self.untrimmed:
                                seq_length = len(read.seq)
                            else:
                                seq_length = read.seq_length_with_start_end_adapters_trimmed()
                            barcode_base_counts[barcode_name] += seq_length

                        # Output all reads to stdin.
                        elif self.output is None:
                            print(read_str, end='')

                        # Output all reads to file.
                        else:
                            out_file.write(read_str)

            # Barcoding summary
            if self.barcode_stats_csv is not None:
                custom_output.close()
            if self.barcode_dir is not None:
                if gzipped_out:
                    logger.info("Compressing output")

                table = [['Barcode', 'Reads', 'Bases', 'File']]
                for barcode_name in sorted(barcode_files.keys()):
                    barcode_files[barcode_name].close()
                    bin_filename = self.barcode_dir / f'{barcode_name}.{self.out_format}'

                    if gzipped_out:
                        if not bin_filename.is_file():
                            continue
                        bin_filename_gz = bin_filename.with_name(bin_filename.name + ".gz")
                        if bin_filename_gz.is_file():
                            bin_filename_gz.unlink()

                        try:
                            subprocess.check_output(f'{gzip_command} {bin_filename}',
                                                    stderr=subprocess.STDOUT, shell=True)
                        except subprocess.CalledProcessError:
                            pass
                        bin_filename = bin_filename_gz

                    table_row = [barcode_name, int_to_str(barcode_read_counts[barcode_name]),
                                int_to_str(barcode_base_counts[barcode_name]), str(bin_filename)]
                    table.append(table_row)

                sys.stderr.write('\n')
                print_table(table, sys.stderr, alignments='LRRL', max_col_width=60, col_separation=2)
                sys.stderr.write('\n')

            elif self.output is None:
                pass

            else:
                out_file.close()
                if gzipped_out:
                    logger.info("Compressing output")
                    subprocess.check_output(f'{gzip_command} -c {out_filename} > {self.output}',
                                            stderr=subprocess.STDOUT, shell=True)
                    out_filename.unlink()
                logger.info(f'Saved result to {self.output}')

            logger.info("Finished trimming")

            # Trimming and splitting summary
            sys.stderr.write('\n' + int_to_str(start_trim_count).rjust(len(int_to_str(total_reads))) +
                ' / ' + int_to_str(total_reads) + ' reads had adapters trimmed from their start (' +
                int_to_str(start_trim_total) + ' bp removed)\n')
            sys.stderr.write(int_to_str(end_trim_count).rjust(len(int_to_str(total_reads))) + ' / ' +
                int_to_str(total_reads) + ' reads had adapters trimmed from their end (' +
                int_to_str(end_trim_total) + ' bp removed)\n')
            verb = 'discarded' if self.discard_middle else 'split'
            sys.stderr.write(int_to_str(middle_trim_count).rjust(len(int_to_str(total_reads))) +
                             ' / ' + int_to_str(total_reads) + ' reads were ' +
                             verb + ' based on middle adapters\n\n')

        except Exception as e:
            # Exit process if exception is caught
            pid = os.getpid()
            self.errors.append((pid, format_exc()))
            self.exit.set()
            os._exit(1)

    def join(self):
        """
        Wait for program finish
        """
        self.process_read.join()
        for p in self.pool:
            p.join()
        self.process_write.join()
        self.raise_exec()

    def raise_exec(self):
        """
        Raise exception if error
        """
        if self.exit.is_set():
            for pid, e in self.errors:
                logging.error(f"worker {pid}:\n {e}")
            logging.error("Terminated")
            sys.exit(1)



"""
Copyright 2017 Ryan Wick (rrwick@gmail.com)
https://github.com/rrwick/Porechop
This module contains the class and sequences for known adapters used in Oxford Nanopore library
preparation kits.
This file is part of Porechop. Porechop is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Porechop is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Porechop. If
not, see <http://www.gnu.org/licenses/>.
"""


class Adapter(object):

    def __init__(self, name, start_sequence=None, end_sequence=None, both_ends_sequence=None):
        self.name = name
        self.start_sequence = start_sequence if start_sequence else []
        self.end_sequence = end_sequence if end_sequence else []
        if both_ends_sequence:
            self.start_sequence = both_ends_sequence
            self.end_sequence = both_ends_sequence
        self.best_start_score, self.best_end_score = 0.0, 0.0

    def best_start_or_end_score(self):
        return max(self.best_start_score, self.best_end_score)

    def is_barcode(self):
        return 'Barcode ' in self.name

    def barcode_direction(self):
        if '_rev' in self.start_sequence[0]:
            return 'reverse'
        else:
            return 'forward'

    def get_barcode_name(self):
        """
        Gets the barcode name for the output files. We want a concise name, so it looks at all
        options and chooses the shortest.
        """
        possible_names = [self.name, self.start_sequence[0]]
        if self.end_sequence:
            possible_names.append(self.end_sequence[0])
        barcode_name = sorted(possible_names, key=lambda x: len(x))[0]
        return barcode_name.replace(' ', '_')


# INSTRUCTIONS FOR ADDING CUSTOM ADAPTERS
# ---------------------------------------
# If you need Porechop to remove adapters that aren't included, you can add your own my modifying
# the ADAPTERS list below.
#
# Here is the format for a normal adapter:
#     Adapter('Adapter_set_name',
#             start_sequence=('Start_adapter_name', 'AAAACCCCGGGGTTTTAAAACCCCGGGGTTTT'),
#             end_sequence=('End_adapter_name', 'AACCGGTTAACCGGTTAACCGGTTAACCGGTT'))
#
# You can exclude start_sequence and end_sequence as appropriate.
#
# If you have custom Barcodes, make sure that the adapter set name starts with 'Barcode '. Also,
# remove the existing barcode sequences from this file to avoid conflicts:
#     Adapter('Barcode 1',
#             start_sequence=('Barcode_1_start', 'AAAAAAAACCCCCCCCGGGGGGGGTTTTTTTT'),
#             end_sequence=('Barcode_1_end', 'AAAAAAAACCCCCCCCGGGGGGGGTTTTTTTT')),
#     Adapter('Barcode 2',
#             start_sequence=('Barcode_2_start', 'TTTTTTTTGGGGGGGGCCCCCCCCAAAAAAAA'),
#             end_sequence=('Barcode_2_end', 'TTTTTTTTGGGGGGGGCCCCCCCCAAAAAAAA'))

NATIVE_BARCODES = [
    # The native barcodes use the rev comp barcode at the start
    # of the read and the forward barcode at the end of the read.
    Adapter('Native Barcode 1',
            start_sequence=('NB01_rev', 'CACAAAGACACCGACAACTTTCTT'),
            end_sequence=('NB01', 'AAGAAAGTTGTCGGTGTCTTTGTG')),
    Adapter('Native Barcode 2',
            start_sequence=('NB02_rev', 'ACAGACGACTACAAACGGAATCGA'),
            end_sequence=('NB02', 'TCGATTCCGTTTGTAGTCGTCTGT')),
    Adapter('Native Barcode 3',
            start_sequence=('NB03_rev', 'CCTGGTAACTGGGACACAAGACTC'),
            end_sequence=('NB03', 'GAGTCTTGTGTCCCAGTTACCAGG')),
    Adapter('Native Barcode 4',
            start_sequence=('NB04_rev', 'TAGGGAAACACGATAGAATCCGAA'),
            end_sequence=('NB04', 'TTCGGATTCTATCGTGTTTCCCTA')),
    Adapter('Native Barcode 5',
            start_sequence=('NB05_rev', 'AAGGTTACACAAACCCTGGACAAG'),
            end_sequence=('NB05', 'CTTGTCCAGGGTTTGTGTAACCTT')),
    Adapter('Native Barcode 6',
            start_sequence=('NB06_rev', 'GACTACTTTCTGCCTTTGCGAGAA'),
            end_sequence=('NB06', 'TTCTCGCAAAGGCAGAAAGTAGTC')),
    Adapter('Native Barcode 7',
            start_sequence=('NB07_rev', 'AAGGATTCATTCCCACGGTAACAC'),
            end_sequence=('NB07', 'GTGTTACCGTGGGAATGAATCCTT')),
    Adapter('Native Barcode 8',
            start_sequence=('NB08_rev', 'ACGTAACTTGGTTTGTTCCCTGAA'),
            end_sequence=('NB08', 'TTCAGGGAACAAACCAAGTTACGT')),
    Adapter('Native Barcode 9',
            start_sequence=('NB09_rev', 'AACCAAGACTCGCTGTGCCTAGTT'),
            end_sequence=('NB09', 'AACTAGGCACAGCGAGTCTTGGTT')),
    Adapter('Native Barcode 10',
            start_sequence=('NB10_rev', 'GAGAGGACAAAGGTTTCAACGCTT'),
            end_sequence=('NB10', 'AAGCGTTGAAACCTTTGTCCTCTC')),
    Adapter('Native Barcode 11',
            start_sequence=('NB11_rev', 'TCCATTCCCTCCGATAGATGAAAC'),
            end_sequence=('NB11', 'GTTTCATCTATCGGAGGGAATGGA')),
    Adapter('Native Barcode 12',
            start_sequence=('NB12_rev', 'TCCGATTCTGCTTCTTTCTACCTG'),
            end_sequence=('NB12', 'CAGGTAGAAAGAAGCAGAATCGGA')),

    Adapter('Native Barcode 13',
            start_sequence=('NB13_rev', 'AGAACGACTTCCATACTCGTGTGA'),
            end_sequence=('NB13', 'TCACACGAGTATGGAAGTCGTTCT')),
    Adapter('Native Barcode 14',
            start_sequence=('NB14_rev', 'AACGAGTCTCTTGGGACCCATAGA'),
            end_sequence=('NB14', 'TCTATGGGTCCCAAGAGACTCGTT')),
    Adapter('Native Barcode 15',
            start_sequence=('NB15_rev', 'AGGTCTACCTCGCTAACACCACTG'),
            end_sequence=('NB15', 'CAGTGGTGTTAGCGAGGTAGACCT')),
    Adapter('Native Barcode 16',
            start_sequence=('NB16_rev', 'CGTCAACTGACAGTGGTTCGTACT'),
            end_sequence=('NB16', 'AGTACGAACCACTGTCAGTTGACG')),
    Adapter('Native Barcode 17',
            start_sequence=('NB17_rev', 'ACCCTCCAGGAAAGTACCTCTGAT'),
            end_sequence=('NB17', 'ATCAGAGGTACTTTCCTGGAGGGT')),
    Adapter('Native Barcode 18',
            start_sequence=('NB18_rev', 'CCAAACCCAACAACCTAGATAGGC'),
            end_sequence=('NB18', 'GCCTATCTAGGTTGTTGGGTTTGG')),
    Adapter('Native Barcode 19',
            start_sequence=('NB19_rev', 'GTTCCTCGTGCAGTGTCAAGAGAT'),
            end_sequence=('NB19', 'ATCTCTTGACACTGCACGAGGAAC')),
    Adapter('Native Barcode 20',
            start_sequence=('NB20_rev', 'TTGCGTCCTGTTACGAGAACTCAT'),
            end_sequence=('NB20', 'ATGAGTTCTCGTAACAGGACGCAA')),
    Adapter('Native Barcode 21',
            start_sequence=('NB21_rev', 'GAGCCTCTCATTGTCCGTTCTCTA'),
            end_sequence=('NB21', 'TAGAGAACGGACAATGAGAGGCTC')),
    Adapter('Native Barcode 22',
            start_sequence=('NB22_rev', 'ACCACTGCCATGTATCAAAGTACG'),
            end_sequence=('NB22', 'CGTACTTTGATACATGGCAGTGGT')),
    Adapter('Native Barcode 23',
            start_sequence=('NB23_rev', 'CTTACTACCCAGTGAACCTCCTCG'),
            end_sequence=('NB23', 'CGAGGAGGTTCACTGGGTAGTAAG')),
    Adapter('Native Barcode 24',
            start_sequence=('NB24_rev', 'GCATAGTTCTGCATGATGGGTTAG'),
            end_sequence=('NB24', 'CTAACCCATCATGCAGAACTATGC'))]

PCR_BARCODES = [
    # Other barcoding kits (like the PCR and rapid barcodes) use the forward barcode at the
    # start of the read and the rev comp barcode at the end of the read.
    Adapter('Barcode 1 (forward)',
            start_sequence=('BC01', 'AAGAAAGTTGTCGGTGTCTTTGTG'),
            end_sequence=('BC01_rev', 'CACAAAGACACCGACAACTTTCTT')),
    Adapter('Barcode 2 (forward)',
            start_sequence=('BC02', 'TCGATTCCGTTTGTAGTCGTCTGT'),
            end_sequence=('BC02_rev', 'ACAGACGACTACAAACGGAATCGA')),
    Adapter('Barcode 3 (forward)',
            start_sequence=('BC03', 'GAGTCTTGTGTCCCAGTTACCAGG'),
            end_sequence=('BC03_rev', 'CCTGGTAACTGGGACACAAGACTC')),
    Adapter('Barcode 4 (forward)',
            start_sequence=('BC04', 'TTCGGATTCTATCGTGTTTCCCTA'),
            end_sequence=('BC04_rev', 'TAGGGAAACACGATAGAATCCGAA')),
    Adapter('Barcode 5 (forward)',
            start_sequence=('BC05', 'CTTGTCCAGGGTTTGTGTAACCTT'),
            end_sequence=('BC05_rev', 'AAGGTTACACAAACCCTGGACAAG')),
    Adapter('Barcode 6 (forward)',
            start_sequence=('BC06', 'TTCTCGCAAAGGCAGAAAGTAGTC'),
            end_sequence=('BC06_rev', 'GACTACTTTCTGCCTTTGCGAGAA')),
    Adapter('Barcode 7 (forward)',
            start_sequence=('BC07', 'GTGTTACCGTGGGAATGAATCCTT'),
            end_sequence=('BC07_rev', 'AAGGATTCATTCCCACGGTAACAC')),
    Adapter('Barcode 8 (forward)',
            start_sequence=('BC08', 'TTCAGGGAACAAACCAAGTTACGT'),
            end_sequence=('BC08_rev', 'ACGTAACTTGGTTTGTTCCCTGAA')),
    Adapter('Barcode 9 (forward)',
            start_sequence=('BC09', 'AACTAGGCACAGCGAGTCTTGGTT'),
            end_sequence=('BC09_rev', 'AACCAAGACTCGCTGTGCCTAGTT')),
    Adapter('Barcode 10 (forward)',
            start_sequence=('BC10', 'AAGCGTTGAAACCTTTGTCCTCTC'),
            end_sequence=('BC10_rev', 'GAGAGGACAAAGGTTTCAACGCTT')),
    Adapter('Barcode 11 (forward)',
            start_sequence=('BC11', 'GTTTCATCTATCGGAGGGAATGGA'),
            end_sequence=('BC11_rev', 'TCCATTCCCTCCGATAGATGAAAC')),
    Adapter('Barcode 12 (forward)',
            start_sequence=('BC12', 'CAGGTAGAAAGAAGCAGAATCGGA'),
            end_sequence=('BC12_rev', 'TCCGATTCTGCTTCTTTCTACCTG')),
    Adapter('Barcode 13 (forward)',
            start_sequence=('BC13', 'AGAACGACTTCCATACTCGTGTGA'),
            end_sequence=('BC13_rev', 'TCACACGAGTATGGAAGTCGTTCT')),
    Adapter('Barcode 14 (forward)',
            start_sequence=('BC14', 'AACGAGTCTCTTGGGACCCATAGA'),
            end_sequence=('BC14_rev', 'TCTATGGGTCCCAAGAGACTCGTT')),
    Adapter('Barcode 15 (forward)',
            start_sequence=('BC15', 'AGGTCTACCTCGCTAACACCACTG'),
            end_sequence=('BC15_rev', 'CAGTGGTGTTAGCGAGGTAGACCT')),
    Adapter('Barcode 16 (forward)',
            start_sequence=('BC16', 'CGTCAACTGACAGTGGTTCGTACT'),
            end_sequence=('BC16_rev', 'AGTACGAACCACTGTCAGTTGACG')),
    Adapter('Barcode 17 (forward)',
            start_sequence=('BC17', 'ACCCTCCAGGAAAGTACCTCTGAT'),
            end_sequence=('BC17_rev', 'ATCAGAGGTACTTTCCTGGAGGGT')),
    Adapter('Barcode 18 (forward)',
            start_sequence=('BC18', 'CCAAACCCAACAACCTAGATAGGC'),
            end_sequence=('BC18_rev', 'GCCTATCTAGGTTGTTGGGTTTGG')),
    Adapter('Barcode 19 (forward)',
            start_sequence=('BC19', 'GTTCCTCGTGCAGTGTCAAGAGAT'),
            end_sequence=('BC19_rev', 'ATCTCTTGACACTGCACGAGGAAC')),
    Adapter('Barcode 20 (forward)',
            start_sequence=('BC20', 'TTGCGTCCTGTTACGAGAACTCAT'),
            end_sequence=('BC20_rev', 'ATGAGTTCTCGTAACAGGACGCAA')),
    Adapter('Barcode 21 (forward)',
            start_sequence=('BC21', 'GAGCCTCTCATTGTCCGTTCTCTA'),
            end_sequence=('BC21_rev', 'TAGAGAACGGACAATGAGAGGCTC')),
    Adapter('Barcode 22 (forward)',
            start_sequence=('BC22', 'ACCACTGCCATGTATCAAAGTACG'),
            end_sequence=('BC22_rev', 'CGTACTTTGATACATGGCAGTGGT')),
    Adapter('Barcode 23 (forward)',
            start_sequence=('BC23', 'CTTACTACCCAGTGAACCTCCTCG'),
            end_sequence=('BC23_rev', 'CGAGGAGGTTCACTGGGTAGTAAG')),
    Adapter('Barcode 24 (forward)',
            start_sequence=('BC24', 'GCATAGTTCTGCATGATGGGTTAG'),
            end_sequence=('BC24_rev', 'CTAACCCATCATGCAGAACTATGC')),
    Adapter('Barcode 25 (forward)',
            start_sequence=('BC25', 'GTAAGTTGGGTATGCAACGCAATG'),
            end_sequence=('BC25_rev', 'CATTGCGTTGCATACCCAACTTAC')),
    Adapter('Barcode 26 (forward)',
            start_sequence=('BC26', 'CATACAGCGACTACGCATTCTCAT'),
            end_sequence=('BC26_rev', 'ATGAGAATGCGTAGTCGCTGTATG')),
    Adapter('Barcode 27 (forward)',
            start_sequence=('BC27', 'CGACGGTTAGATTCACCTCTTACA'),
            end_sequence=('BC27_rev', 'TGTAAGAGGTGAATCTAACCGTCG')),
    Adapter('Barcode 28 (forward)',
            start_sequence=('BC28', 'TGAAACCTAAGAAGGCACCGTATC'),
            end_sequence=('BC28_rev', 'GATACGGTGCCTTCTTAGGTTTCA')),
    Adapter('Barcode 29 (forward)',
            start_sequence=('BC29', 'CTAGACACCTTGGGTTGACAGACC'),
            end_sequence=('BC29_rev', 'GGTCTGTCAACCCAAGGTGTCTAG')),
    Adapter('Barcode 30 (forward)',
            start_sequence=('BC30', 'TCAGTGAGGATCTACTTCGACCCA'),
            end_sequence=('BC30_rev', 'TGGGTCGAAGTAGATCCTCACTGA')),
    Adapter('Barcode 31 (forward)',
            start_sequence=('BC31', 'TGCGTACAGCAATCAGTTACATTG'),
            end_sequence=('BC31_rev', 'CAATGTAACTGATTGCTGTACGCA')),
    Adapter('Barcode 32 (forward)',
            start_sequence=('BC32', 'CCAGTAGAAGTCCGACAACGTCAT'),
            end_sequence=('BC32_rev', 'ATGACGTTGTCGGACTTCTACTGG')),
    Adapter('Barcode 33 (forward)',
            start_sequence=('BC33', 'CAGACTTGGTACGGTTGGGTAACT'),
            end_sequence=('BC33_rev', 'AGTTACCCAACCGTACCAAGTCTG')),
    Adapter('Barcode 34 (forward)',
            start_sequence=('BC34', 'GGACGAAGAACTCAAGTCAAAGGC'),
            end_sequence=('BC34_rev', 'GCCTTTGACTTGAGTTCTTCGTCC')),
    Adapter('Barcode 35 (forward)',
            start_sequence=('BC35', 'CTACTTACGAAGCTGAGGGACTGC'),
            end_sequence=('BC35_rev', 'GCAGTCCCTCAGCTTCGTAAGTAG')),
    Adapter('Barcode 36 (forward)',
            start_sequence=('BC36', 'ATGTCCCAGTTAGAGGAGGAAACA'),
            end_sequence=('BC36_rev', 'TGTTTCCTCCTCTAACTGGGACAT')),
    Adapter('Barcode 37 (forward)',
            start_sequence=('BC37', 'GCTTGCGATTGATGCTTAGTATCA'),
            end_sequence=('BC37_rev', 'TGATACTAAGCATCAATCGCAAGC')),
    Adapter('Barcode 38 (forward)',
            start_sequence=('BC38', 'ACCACAGGAGGACGATACAGAGAA'),
            end_sequence=('BC38_rev', 'TTCTCTGTATCGTCCTCCTGTGGT')),
    Adapter('Barcode 39 (forward)',
            start_sequence=('BC39', 'CCACAGTGTCAACTAGAGCCTCTC'),
            end_sequence=('BC39_rev', 'GAGAGGCTCTAGTTGACACTGTGG')),
    Adapter('Barcode 40 (forward)',
            start_sequence=('BC40', 'TAGTTTGGATGACCAAGGATAGCC'),
            end_sequence=('BC40_rev', 'GGCTATCCTTGGTCATCCAAACTA')),
    Adapter('Barcode 41 (forward)',
            start_sequence=('BC41', 'GGAGTTCGTCCAGAGAAGTACACG'),
            end_sequence=('BC41_rev', 'CGTGTACTTCTCTGGACGAACTCC')),
    Adapter('Barcode 42 (forward)',
            start_sequence=('BC42', 'CTACGTGTAAGGCATACCTGCCAG'),
            end_sequence=('BC42_rev', 'CTGGCAGGTATGCCTTACACGTAG')),
    Adapter('Barcode 43 (forward)',
            start_sequence=('BC43', 'CTTTCGTTGTTGACTCGACGGTAG'),
            end_sequence=('BC43_rev', 'CTACCGTCGAGTCAACAACGAAAG')),
    Adapter('Barcode 44 (forward)',
            start_sequence=('BC44', 'AGTAGAAAGGGTTCCTTCCCACTC'),
            end_sequence=('BC44_rev', 'GAGTGGGAAGGAACCCTTTCTACT')),
    Adapter('Barcode 45 (forward)',
            start_sequence=('BC45', 'GATCCAACAGAGATGCCTTCAGTG'),
            end_sequence=('BC45_rev', 'CACTGAAGGCATCTCTGTTGGATC')),
    Adapter('Barcode 46 (forward)',
            start_sequence=('BC46', 'GCTGTGTTCCACTTCATTCTCCTG'),
            end_sequence=('BC46_rev', 'CAGGAGAATGAAGTGGAACACAGC')),
    Adapter('Barcode 47 (forward)',
            start_sequence=('BC47', 'GTGCAACTTTCCCACAGGTAGTTC'),
            end_sequence=('BC47_rev', 'GAACTACCTGTGGGAAAGTTGCAC')),
    Adapter('Barcode 48 (forward)',
            start_sequence=('BC48', 'CATCTGGAACGTGGTACACCTGTA'),
            end_sequence=('BC48_rev', 'TACAGGTGTACCACGTTCCAGATG')),
    Adapter('Barcode 49 (forward)',
            start_sequence=('BC49', 'ACTGGTGCAGCTTTGAACATCTAG'),
            end_sequence=('BC49_rev', 'CTAGATGTTCAAAGCTGCACCAGT')),
    Adapter('Barcode 50 (forward)',
            start_sequence=('BC50', 'ATGGACTTTGGTAACTTCCTGCGT'),
            end_sequence=('BC50_rev', 'ACGCAGGAAGTTACCAAAGTCCAT')),
    Adapter('Barcode 51 (forward)',
            start_sequence=('BC51', 'GTTGAATGAGCCTACTGGGTCCTC'),
            end_sequence=('BC51_rev', 'GAGGACCCAGTAGGCTCATTCAAC')),
    Adapter('Barcode 52 (forward)',
            start_sequence=('BC52', 'TGAGAGACAAGATTGTTCGTGGAC'),
            end_sequence=('BC52_rev', 'GTCCACGAACAATCTTGTCTCTCA')),
    Adapter('Barcode 53 (forward)',
            start_sequence=('BC53', 'AGATTCAGACCGTCTCATGCAAAG'),
            end_sequence=('BC53_rev', 'CTTTGCATGAGACGGTCTGAATCT')),
    Adapter('Barcode 54 (forward)',
            start_sequence=('BC54', 'CAAGAGCTTTGACTAAGGAGCATG'),
            end_sequence=('BC54_rev', 'CATGCTCCTTAGTCAAAGCTCTTG')),
    Adapter('Barcode 55 (forward)',
            start_sequence=('BC55', 'TGGAAGATGAGACCCTGATCTACG'),
            end_sequence=('BC55_rev', 'CGTAGATCAGGGTCTCATCTTCCA')),
    Adapter('Barcode 56 (forward)',
            start_sequence=('BC56', 'TCACTACTCAACAGGTGGCATGAA'),
            end_sequence=('BC56_rev', 'TTCATGCCACCTGTTGAGTAGTGA')),
    Adapter('Barcode 57 (forward)',
            start_sequence=('BC57', 'GCTAGGTCAATCTCCTTCGGAAGT'),
            end_sequence=('BC57_rev', 'ACTTCCGAAGGAGATTGACCTAGC')),
    Adapter('Barcode 58 (forward)',
            start_sequence=('BC58', 'CAGGTTACTCCTCCGTGAGTCTGA'),
            end_sequence=('BC58_rev', 'TCAGACTCACGGAGGAGTAACCTG')),
    Adapter('Barcode 59 (forward)',
            start_sequence=('BC59', 'TCAATCAAGAAGGGAAAGCAAGGT'),
            end_sequence=('BC59_rev', 'ACCTTGCTTTCCCTTCTTGATTGA')),
    Adapter('Barcode 60 (forward)',
            start_sequence=('BC60', 'CATGTTCAACCAAGGCTTCTATGG'),
            end_sequence=('BC60_rev', 'CCATAGAAGCCTTGGTTGAACATG')),
    Adapter('Barcode 61 (forward)',
            start_sequence=('BC61', 'AGAGGGTACTATGTGCCTCAGCAC'),
            end_sequence=('BC61_rev', 'GTGCTGAGGCACATAGTACCCTCT')),
    Adapter('Barcode 62 (forward)',
            start_sequence=('BC62', 'CACCCACACTTACTTCAGGACGTA'),
            end_sequence=('BC62_rev', 'TACGTCCTGAAGTAAGTGTGGGTG')),
    Adapter('Barcode 63 (forward)',
            start_sequence=('BC63', 'TTCTGAAGTTCCTGGGTCTTGAAC'),
            end_sequence=('BC63_rev', 'GTTCAAGACCCAGGAACTTCAGAA')),
    Adapter('Barcode 64 (forward)',
            start_sequence=('BC64', 'GACAGACACCGTTCATCGACTTTC'),
            end_sequence=('BC64_rev', 'GAAAGTCGATGAACGGTGTCTGTC')),
    Adapter('Barcode 65 (forward)',
            start_sequence=('BC65', 'TTCTCAGTCTTCCTCCAGACAAGG'),
            end_sequence=('BC65_rev', 'CCTTGTCTGGAGGAAGACTGAGAA')),
    Adapter('Barcode 66 (forward)',
            start_sequence=('BC66', 'CCGATCCTTGTGGCTTCTAACTTC'),
            end_sequence=('BC66_rev', 'GAAGTTAGAAGCCACAAGGATCGG')),
    Adapter('Barcode 67 (forward)',
            start_sequence=('BC67', 'GTTTGTCATACTCGTGTGCTCACC'),
            end_sequence=('BC67_rev', 'GGTGAGCACACGAGTATGACAAAC')),
    Adapter('Barcode 68 (forward)',
            start_sequence=('BC68', 'GAATCTAAGCAAACACGAAGGTGG'),
            end_sequence=('BC68_rev', 'CCACCTTCGTGTTTGCTTAGATTC')),
    Adapter('Barcode 69 (forward)',
            start_sequence=('BC69', 'TACAGTCCGAGCCTCATGTGATCT'),
            end_sequence=('BC69_rev', 'AGATCACATGAGGCTCGGACTGTA')),
    Adapter('Barcode 70 (forward)',
            start_sequence=('BC70', 'ACCGAGATCCTACGAATGGAGTGT'),
            end_sequence=('BC70_rev', 'ACACTCCATTCGTAGGATCTCGGT')),
    Adapter('Barcode 71 (forward)',
            start_sequence=('BC71', 'CCTGGGAGCATCAGGTAGTAACAG'),
            end_sequence=('BC71_rev', 'CTGTTACTACCTGATGCTCCCAGG')),
    Adapter('Barcode 72 (forward)',
            start_sequence=('BC72', 'TAGCTGACTGTCTTCCATACCGAC'),
            end_sequence=('BC72_rev', 'GTCGGTATGGAAGACAGTCAGCTA')),
    Adapter('Barcode 73 (forward)',
            start_sequence=('BC73', 'AAGAAACAGGATGACAGAACCCTC'),
            end_sequence=('BC73_rev', 'GAGGGTTCTGTCATCCTGTTTCTT')),
    Adapter('Barcode 74 (forward)',
            start_sequence=('BC74', 'TACAAGCATCCCAACACTTCCACT'),
            end_sequence=('BC74_rev', 'AGTGGAAGTGTTGGGATGCTTGTA')),
    Adapter('Barcode 75 (forward)',
            start_sequence=('BC75', 'GACCATTGTGATGAACCCTGTTGT'),
            end_sequence=('BC75_rev', 'ACAACAGGGTTCATCACAATGGTC')),
    Adapter('Barcode 76 (forward)',
            start_sequence=('BC76', 'ATGCTTGTTACATCAACCCTGGAC'),
            end_sequence=('BC76_rev', 'GTCCAGGGTTGATGTAACAAGCAT')),
    Adapter('Barcode 77 (forward)',
            start_sequence=('BC77', 'CGACCTGTTTCTCAGGGATACAAC'),
            end_sequence=('BC77_rev', 'GTTGTATCCCTGAGAAACAGGTCG')),
    Adapter('Barcode 78 (forward)',
            start_sequence=('BC78', 'AACAACCGAACCTTTGAATCAGAA'),
            end_sequence=('BC78_rev', 'TTCTGATTCAAAGGTTCGGTTGTT')),
    Adapter('Barcode 79 (forward)',
            start_sequence=('BC79', 'TCTCGGAGATAGTTCTCACTGCTG'),
            end_sequence=('BC79_rev', 'CAGCAGTGAGAACTATCTCCGAGA')),
    Adapter('Barcode 80 (forward)',
            start_sequence=('BC80', 'CGGATGAACATAGGATAGCGATTC'),
            end_sequence=('BC80_rev', 'GAATCGCTATCCTATGTTCATCCG')),
    Adapter('Barcode 81 (forward)',
            start_sequence=('BC81', 'CCTCATCTTGTGAAGTTGTTTCGG'),
            end_sequence=('BC81_rev', 'CCGAAACAACTTCACAAGATGAGG')),
    Adapter('Barcode 82 (forward)',
            start_sequence=('BC82', 'ACGGTATGTCGAGTTCCAGGACTA'),
            end_sequence=('BC82_rev', 'TAGTCCTGGAACTCGACATACCGT')),
    Adapter('Barcode 83 (forward)',
            start_sequence=('BC83', 'TGGCTTGATCTAGGTAAGGTCGAA'),
            end_sequence=('BC83_rev', 'TTCGACCTTACCTAGATCAAGCCA')),
    Adapter('Barcode 84 (forward)',
            start_sequence=('BC84', 'GTAGTGGACCTAGAACCTGTGCCA'),
            end_sequence=('BC84_rev', 'TGGCACAGGTTCTAGGTCCACTAC')),
    Adapter('Barcode 85 (forward)',
            start_sequence=('BC85', 'AACGGAGGAGTTAGTTGGATGATC'),
            end_sequence=('BC85_rev', 'GATCATCCAACTAACTCCTCCGTT')),
    Adapter('Barcode 86 (forward)',
            start_sequence=('BC86', 'AGGTGATCCCAACAAGCGTAAGTA'),
            end_sequence=('BC86_rev', 'TACTTACGCTTGTTGGGATCACCT')),
    Adapter('Barcode 87 (forward)',
            start_sequence=('BC87', 'TACATGCTCCTGTTGTTAGGGAGG'),
            end_sequence=('BC87_rev', 'CCTCCCTAACAACAGGAGCATGTA')),
    Adapter('Barcode 88 (forward)',
            start_sequence=('BC88', 'TCTTCTACTACCGATCCGAAGCAG'),
            end_sequence=('BC88_rev', 'CTGCTTCGGATCGGTAGTAGAAGA')),
    Adapter('Barcode 89 (forward)',
            start_sequence=('BC89', 'ACAGCATCAATGTTTGGCTAGTTG'),
            end_sequence=('BC89_rev', 'CAACTAGCCAAACATTGATGCTGT')),
    Adapter('Barcode 90 (forward)',
            start_sequence=('BC90', 'GATGTAGAGGGTACGGTTTGAGGC'),
            end_sequence=('BC90_rev', 'GCCTCAAACCGTACCCTCTACATC')),
    Adapter('Barcode 91 (forward)',
            start_sequence=('BC91', 'GGCTCCATAGGAACTCACGCTACT'),
            end_sequence=('BC91_rev', 'AGTAGCGTGAGTTCCTATGGAGCC')),
    Adapter('Barcode 92 (forward)',
            start_sequence=('BC92', 'TTGTGAGTGGAAAGATACAGGACC'),
            end_sequence=('BC92_rev', 'GGTCCTGTATCTTTCCACTCACAA')),
    Adapter('Barcode 93 (forward)',
            start_sequence=('BC93', 'AGTTTCCATCACTTCAGACTTGGG'),
            end_sequence=('BC93_rev', 'CCCAAGTCTGAAGTGATGGAAACT')),
    Adapter('Barcode 94 (forward)',
            start_sequence=('BC94', 'GATTGTCCTCAAACTGCCACCTAC'),
            end_sequence=('BC94_rev', 'GTAGGTGGCAGTTTGAGGACAATC')),
    Adapter('Barcode 95 (forward)',
            start_sequence=('BC95', 'CCTGTCTGGAAGAAGAATGGACTT'),
            end_sequence=('BC95_rev', 'AAGTCCATTCTTCTTCCAGACAGG')),
    Adapter('Barcode 96 (forward)',
            start_sequence=('BC96', 'CTGAACGGTCATAGAGTCCACCAT'),
            end_sequence=('BC96_rev', 'ATGGTGGACTCTATGACCGTTCAG'))]

RAPID_BARCODES = [
    Adapter('Rapid Barcode 1 (reverse)',
            start_sequence=('BC01', 'AAGAAAGTTGTCGGTGTCTTTGTG')),
    Adapter('Rapid Barcode 2 (reverse)',
            start_sequence=('BC02', 'TCGATTCCGTTTGTAGTCGTCTGT')),
    Adapter('Rapid Barcode 3 (reverse)',
            start_sequence=('BC03', 'GAGTCTTGTGTCCCAGTTACCAGG')),
    Adapter('Rapid Barcode 4 (reverse)',
            start_sequence=('BC04', 'TTCGGATTCTATCGTGTTTCCCTA')),
    Adapter('Rapid Barcode 5 (reverse)',
            start_sequence=('BC05', 'CTTGTCCAGGGTTTGTGTAACCTT')),
    Adapter('Rapid Barcode 6 (reverse)',
            start_sequence=('BC06', 'TTCTCGCAAAGGCAGAAAGTAGTC')),
    Adapter('Rapid Barcode 7 (reverse)',
            start_sequence=('BC07', 'GTGTTACCGTGGGAATGAATCCTT')),
    Adapter('Rapid Barcode 8 (reverse)',
            start_sequence=('BC08', 'TTCAGGGAACAAACCAAGTTACGT')),
    Adapter('Rapid Barcode 9 (reverse)',
            start_sequence=('BC09', 'AACTAGGCACAGCGAGTCTTGGTT')),
    Adapter('Rapid Barcode 10 (reverse)',
            start_sequence=('BC10', 'AAGCGTTGAAACCTTTGTCCTCTC')),
    Adapter('Rapid Barcode 11 (reverse)',
            start_sequence=('BC11', 'GTTTCATCTATCGGAGGGAATGGA')),
    Adapter('Rapid Barcode 12 (reverse)',
            start_sequence=('BC12', 'CAGGTAGAAAGAAGCAGAATCGGA')),
]

OTHER_ADAPTERS = [
    Adapter('Barcode 13 (reverse)',
            start_sequence=('BC13', 'AGAACGACTTCCATACTCGTGTGA'),
            end_sequence=('BC13_rev', 'TCACACGAGTATGGAAGTCGTTCT')),
    Adapter('Barcode 14 (reverse)',
            start_sequence=('BC14', 'AACGAGTCTCTTGGGACCCATAGA'),
            end_sequence=('BC14_rev', 'TCTATGGGTCCCAAGAGACTCGTT')),
    Adapter('Barcode 15 (reverse)',
            start_sequence=('BC15', 'AGGTCTACCTCGCTAACACCACTG'),
            end_sequence=('BC15_rev', 'CAGTGGTGTTAGCGAGGTAGACCT')),
    Adapter('Barcode 16 (reverse)',
            start_sequence=('BC16', 'CGTCAACTGACAGTGGTTCGTACT'),
            end_sequence=('BC16_rev', 'AGTACGAACCACTGTCAGTTGACG')),
    Adapter('Barcode 17 (reverse)',
            start_sequence=('BC17', 'ACCCTCCAGGAAAGTACCTCTGAT'),
            end_sequence=('BC17_rev', 'ATCAGAGGTACTTTCCTGGAGGGT')),
    Adapter('Barcode 18 (reverse)',
            start_sequence=('BC18', 'CCAAACCCAACAACCTAGATAGGC'),
            end_sequence=('BC18_rev', 'GCCTATCTAGGTTGTTGGGTTTGG')),
    Adapter('Barcode 19 (reverse)',
            start_sequence=('BC19', 'GTTCCTCGTGCAGTGTCAAGAGAT'),
            end_sequence=('BC19_rev', 'ATCTCTTGACACTGCACGAGGAAC')),
    Adapter('Barcode 20 (reverse)',
            start_sequence=('BC20', 'TTGCGTCCTGTTACGAGAACTCAT'),
            end_sequence=('BC20_rev', 'ATGAGTTCTCGTAACAGGACGCAA')),
    Adapter('Barcode 21 (reverse)',
            start_sequence=('BC21', 'GAGCCTCTCATTGTCCGTTCTCTA'),
            end_sequence=('BC21_rev', 'TAGAGAACGGACAATGAGAGGCTC')),
    Adapter('Barcode 22 (reverse)',
            start_sequence=('BC22', 'ACCACTGCCATGTATCAAAGTACG'),
            end_sequence=('BC22_rev', 'CGTACTTTGATACATGGCAGTGGT')),
    Adapter('Barcode 23 (reverse)',
            start_sequence=('BC23', 'CTTACTACCCAGTGAACCTCCTCG'),
            end_sequence=('BC23_rev', 'CGAGGAGGTTCACTGGGTAGTAAG')),
    Adapter('Barcode 24 (reverse)',
            start_sequence=('BC24', 'GCATAGTTCTGCATGATGGGTTAG'),
            end_sequence=('BC24_rev', 'CTAACCCATCATGCAGAACTATGC')),
    Adapter('SQK-NSK007',
            start_sequence=('SQK-NSK007_Y_Top', 'AATGTACTTCGTTCAGTTACGTATTGCT'),
            end_sequence=('SQK-NSK007_Y_Bottom', 'GCAATACGTAACTGAACGAAGT')),

    Adapter('Rapid',
            start_sequence=('Rapid_adapter', 'GTTTTCGCATTTATCGTGAAACGCTTTCGCGTTTTTCGTGCGCCGCTTCA')),

    Adapter('SQK-MAP006',
            start_sequence=('SQK-MAP006_Y_Top_SK63',    'GGTTGTTTCTGTTGGTGCTGATATTGCT'),
            end_sequence=  ('SQK-MAP006_Y_Bottom_SK64', 'GCAATATCAGCACCAACAGAAA')),

    Adapter('SQK-MAP006 Short',
            start_sequence=('SQK-MAP006_Short_Y_Top_LI32',    'CGGCGTCTGCTTGGGTGTTTAACCT'),
            end_sequence=  ('SQK-MAP006_Short_Y_Bottom_LI33', 'GGTTAAACACCCAAGCAGACGCCG')),

    Adapter('PCR adapters 1',
            start_sequence=('PCR_1_start', 'ACTTGCCTGTCGCTCTATCTTC'),
            end_sequence=  ('PCR_1_end', 'GAAGATAGAGCGACAGGCAAGT')),

    Adapter('PCR tail 1',
            start_sequence=('PCR_tail_1_start', 'TTAACCTTTCTGTTGGTGCTGATATTGC'),
            end_sequence=  ('PCR_tail_1_end',   'GCAATATCAGCACCAACAGAAAGGTTAA')),

    Adapter('PCR tail 2',
            start_sequence=('PCR_tail_2_start', 'TTAACCTACTTGCCTGTCGCTCTATCTTC'),
            end_sequence=  ('PCR_tail_2_end',   'GAAGATAGAGCGACAGGCAAGTAGGTTAA')),


    # 1D^2 kit adapters are interesting. ONT provided the following sequences on their site:
    #   start: GGCGTCTGCTTGGGTGTTTAACCTTTTTGTCAGAGAGGTTCCAAGTCAGAGAGGTTCCT
    #   end:   GGAACCTCTCTCTGACTTGGAACCTCTCTGACAAAAAGGTTAAACACCCAAGCAGACGCCAGCAAT
    # But when looking at actual reads, I found two parts. The first corresponds to one end
    # of the provided sequences (through slightly different):
    Adapter('1D^2 part 1',
            start_sequence=('1D2_part_1_start', 'GAGAGGTTCCAAGTCAGAGAGGTTCCT'),
            end_sequence=  ('1D2_part_1_end',   'AGGAACCTCTCTGACTTGGAACCTCTC')),
    # and the second part corresponds to the other end, combined with a bit of standard 1D
    # adapter:
    Adapter('1D^2 part 2',
            start_sequence=('1D2_part_2_start', 'CTTCGTTCAGTTACGTATTGCTGGCGTCTGCTT'),
            end_sequence=  ('1D2_part_2_end',   'CACCCAAGCAGACGCCAGCAATACGTAACT')),
    # The middle part of the provided sequences is less common, so I've left it out of the
    # adapter sequences here.
    ]

ALL_ADAPTERS = NATIVE_BARCODES + PCR_BARCODES + RAPID_BARCODES + OTHER_ADAPTERS


def make_full_native_barcode_adapter(barcode_num):
    barcode = [x for x in NATIVE_BARCODES if x.name == 'Native Barcode ' + str(barcode_num)][0]
    start_barcode_seq = barcode.start_sequence[1]
    end_barcode_seq = barcode.end_sequence[1]

    start_full_seq = 'AATGTACTTCGTTCAGTTACGTATTGCTAAGGTTAA' + start_barcode_seq + 'CAGCACCT'
    end_full_seq = 'AGGTGCTG' + end_barcode_seq + 'TTAACCTTAGCAATACGTAACTGAACGAAGT'

    return Adapter('Native barcoding ' + str(barcode_num) + ' (full sequence)',
                   start_sequence=('NB' + '%02d' % barcode_num + '_start', start_full_seq),
                   end_sequence=('NB' + '%02d' % barcode_num + '_end', end_full_seq))


def make_full_rapid_barcode_adapter(barcode_num):
    barcode = [x for x in ALL_ADAPTERS if x.name == 'Barcode ' + str(barcode_num) + ' (forward)'][0]
    start_barcode_seq = barcode.start_sequence[1]

    start_full_seq = 'AATGTACTTCGTTCAGTTACGTATTGCT' + start_barcode_seq + 'GTTTTCGCATTTATCGTGAAACGCTTTCGCGTTTTTCGTGCGCCGCTTCA'

    return Adapter('Rapid barcoding ' + str(barcode_num) + ' (full sequence)',
                   start_sequence=('RB' + '%02d' % barcode_num + '_full', start_full_seq))

def load_primers(primer_file):
    c = 0
    # file should be csv eg:
    # primer1_LEFT,seq\n
    # primer1_right,seq\n
    # primer2_LEFT,seq\n
    # primer2_right,seq\n

    print("Loading primers from file {}...".format(primer_file))
    CUSTOM_BARCODES=[]
    start_name=''
    end_name=''
    name=''
    custom_barcodes = str(primer_file)
    with open(custom_barcodes,"r") as f:
        for l in f:
            c +=1
            tokens=l.split(',')
            if c%2!=0:
                name='Barcode {} (forward)'.format(tokens[0].rstrip('_LEFT'))
                start_name=tokens[0]
                start_seq=tokens[1]
            else:
                if name=='Barcode {} (forward)'.format(tokens[0].rstrip('_RIGHT')):
                    end_name=tokens[0]
                    end_seq=tokens[1]
                    print(name, start_name, end_name, start_seq, end_seq)
                    CUSTOM_BARCODES.append(Adapter(name, start_sequence=(start_name, start_seq), end_sequence=(end_name, end_seq)))
    return CUSTOM_BARCODES

def load_custom_barcodes(custom_barcode_file):
    c = 0
    # file should be csv with columns [Name,Direction{1=Forward,0=Reverse},StartSeq,EndSeq]
    print("Loading custom barcodes from file {}...".format(custom_barcode_file))
    CUSTOM_BARCODES=[]
    start_name=''
    end_name=''
    name=''
    custom_barcodes = str(custom_barcode_file)
    barcode_dict = {}
    with open(custom_barcodes,"r") as f:
        for l in f:
            if l.startswith("#"):
                continue
            c +=1
            tokens=l.split(',')
            name=tokens[0]
            direction = tokens[1]
            if direction:
                ad_name=name + ' (forward)'
                end_name=name+'_rev'
                CUSTOM_BARCODES.append(Adapter(ad_name, start_sequence=(name, tokens[2]), end_sequence=(end_name, tokens[3])))
            else:
                ad_name=name + ' (reverse)'
                start_name=name+'_rev'
                CUSTOM_BARCODES.append(Adapter(ad_name, start_sequence=(start_name, tokens[2]), end_sequence=(name, tokens[3])))
    return CUSTOM_BARCODES



"""
Copyright 2017 Ryan Wick (rrwick@gmail.com)
https://github.com/rrwick/Porechop

This module contains the class for a Nanopore read.

This file is part of Porechop. Porechop is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version. Porechop is distributed in
the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details. You should have received a copy of the GNU General Public License along with Porechop. If
not, see <http://www.gnu.org/licenses/>.
"""
import sys
import edlib
import logging
import parasail
from tqdm import tqdm
from multiprocessing.dummy import Pool as ThreadPool
from porechopx.adapters import make_full_native_barcode_adapter, make_full_rapid_barcode_adapter
from porechopx.misc import yellow, red, add_line_breaks_to_sequence, END_FORMATTING, RED, YELLOW
logger = logging.getLogger("porechopx")


class NanoporeRead(object):

    def __init__(self, name, seq, quals):
        self.name = name

        self.seq = seq
        self.quals = quals
        if len(quals) < len(seq):
            self.quals += '+' * (len(seq) - len(quals))

        self.start_trim_amount = 0
        self.end_trim_amount = 0
        self.start_adapter_alignments = []
        self.end_adapter_alignments = []

        self.middle_adapter_positions = set()
        self.middle_trim_positions = set()
        self.middle_hit_str = ''
        self.middle_hit_id = []

        self.start_barcode_scores = {}
        self.end_barcode_scores = {}

        self.best_start_barcode = ('none', 0.0)
        self.best_end_barcode = ('none', 0.0)
        self.second_best_start_barcode = ('none', 0.0)
        self.second_best_end_barcode = ('none', 0.0)
        self.barcode_call = 'none'
        self.albacore_barcode_call = None

    def get_name(self, barcode_labels=False, extended_labels=False):
        if barcode_labels:
            if extended_labels:
                start_name, start_id, end_name, end_id, middle_name, middle_id = self.get_barcode_hits()
                bcode_info = " barcode={} start_name={} start_id={} end_name={} end_id={}".format(self.barcode_call, start_name, start_id, end_name, end_id)
                if middle_name:
                    return self.name + bcode_info + " middle_name=" + middle_name + " middle_id=" + middle_id
                else:
                    return self.name + bcode_info + " middle_name=none middle_id=none"
            else:
                return self.name + " barcode=" + self.barcode_call

        return self.name

    def get_barcode_hits(self):
        start_name, start_id = self.best_start_barcode
        end_name, end_id = self.best_end_barcode

        if self.middle_hit_id != []:
            best_middle_hit = sorted(self.middle_hit_id, key = lambda x : float(x[1]), reverse=True)[0]

            return start_name, start_id, end_name, end_id, best_middle_hit[0], best_middle_hit[1]

        return start_name, start_id, end_name, end_id, None, None

    def write_to_stats_csv(self, custom_output):
        """
        Write barcoding status into csv:
            start_time has been removed to support both dorado and guppy fastq
        """
        # TODO: unify read information for guppy and dorado basecalled fastq

        name = self.name.split()[0]
        start_name, start_id, end_name, end_id, middle_name, middle_id = self.get_barcode_hits()
        if not middle_name:
            middle_name, middle_id = ("none","none")
        custom_output.write(f"{name},{self.barcode_call},{start_name},{start_id},{end_name},{end_id},{middle_name},{middle_id}\n")

    def get_seq_with_start_end_adapters_trimmed(self):
        if not self.start_trim_amount and not self.end_trim_amount:
            return self.seq
        start_pos = self.start_trim_amount
        end_pos = len(self.seq) - self.end_trim_amount
        trimmed_seq = self.seq[start_pos:end_pos]
        return trimmed_seq

    def seq_length_with_start_end_adapters_trimmed(self):
        return len(self.get_seq_with_start_end_adapters_trimmed())

    def get_quals_with_start_end_adapters_trimmed(self):
        if not self.start_trim_amount and not self.end_trim_amount:
            return self.quals
        start_pos = self.start_trim_amount
        end_pos = len(self.quals) - self.end_trim_amount
        trimmed_quals = self.quals[start_pos:end_pos]
        return trimmed_quals

    def get_split_read_parts(self, min_split_read_size):
        """
        Returns the read split into parts as determined by the middle_trim_positions set.
        """
        trimmed_seq = self.get_seq_with_start_end_adapters_trimmed()
        trimmed_quals = self.get_quals_with_start_end_adapters_trimmed()
        split_read_parts = []
        part_seq, part_quals = [], []
        for i in range(len(trimmed_seq)):
            if i in self.middle_trim_positions:
                if part_seq:
                    split_read_parts.append((''.join(part_seq), ''.join(part_quals)))
                    part_seq, part_quals = [], []
            else:
                part_seq.append(trimmed_seq[i])
                part_quals.append(trimmed_quals[i])
        if part_seq:
            split_read_parts.append((''.join(part_seq), ''.join(part_quals)))
        split_read_parts = [x for x in split_read_parts if len(x[0]) >= min_split_read_size]
        return split_read_parts

    def get_fasta(self, min_split_read_size, discard_middle, untrimmed=False, barcode_labels=False, extended_labels=False):
        if not self.middle_trim_positions:
            if untrimmed:
                seq = self.seq
            else:
                seq = self.get_seq_with_start_end_adapters_trimmed()
            if not seq:  # Don't return empty sequences
                return ''
            return ''.join(['>', self.get_name(barcode_labels, extended_labels), '\n', add_line_breaks_to_sequence(seq, 70)])
        elif discard_middle:
            return ''
        else:
            fasta_str = ''
            for i, split_read_part in enumerate(self.get_split_read_parts(min_split_read_size)):
                read_name = add_number_to_read_name(self.get_name(), i + 1)
                if not split_read_part[0]:  # Don't return empty sequences
                    return ''
                seq = add_line_breaks_to_sequence(split_read_part[0], 70)
                fasta_str += ''.join(['>', read_name, '\n', seq])
            return fasta_str

    def get_fastq(self, min_split_read_size, discard_middle, untrimmed=False, barcode_labels=False, extended_labels=False):
        if not self.middle_trim_positions:
            if untrimmed:
                seq = self.seq
                quals = self.quals
            else:
                seq = self.get_seq_with_start_end_adapters_trimmed()
                quals = self.get_quals_with_start_end_adapters_trimmed()
            if not seq:  # Don't return empty sequences
                return ''
            return ''.join(['@', self.get_name(barcode_labels,extended_labels), '\n', seq, '\n+\n', quals, '\n'])
        elif discard_middle:
            return ''
        else:
            fastq_str = ''
            for i, split_read_part in enumerate(self.get_split_read_parts(min_split_read_size)):
                read_name = add_number_to_read_name(self.get_name(), i + 1)
                if not split_read_part[0]:  # Don't return empty sequences
                    return ''
                fastq_str += ''.join(['@', read_name, '\n', split_read_part[0], '\n+\n',
                                      split_read_part[1], '\n'])
            return fastq_str

    def find_start_trim(self, adapters, end_size, extra_trim_size, end_threshold,
                        scoring_scheme_vals, min_trim_size, check_barcodes, forward_or_reverse):
        """
        Aligns one or more adapter sequences and possibly adjusts the read's start trim amount based
        on the result.
        """
        read_seq_start = self.seq[:end_size]
        for adapter in adapters:
            full_score, partial_score, read_start, read_end = \
                align_adapter(read_seq_start, adapter.start_sequence[1], *scoring_scheme_vals)
            if partial_score > end_threshold and read_end != end_size and \
                    read_end - read_start >= min_trim_size:
                trim_amount = read_end + extra_trim_size
                self.start_trim_amount = max(self.start_trim_amount, trim_amount)
                self.start_adapter_alignments.append((adapter, full_score, partial_score,
                                                      read_start, read_end))
            if check_barcodes and adapter.is_barcode() and \
                    adapter.barcode_direction() == forward_or_reverse:
                self.start_barcode_scores[adapter.get_barcode_name()] = full_score

    def find_end_trim(self, adapters, end_size, extra_trim_size, end_threshold,
                      scoring_scheme_vals, min_trim_size, check_barcodes, forward_or_reverse):
        """
        Aligns one or more adapter sequences and possibly adjusts the read's end trim amount based
        on the result.
        """
        read_seq_end = self.seq[-end_size:]
        for adapter in adapters:
            if not adapter.end_sequence:
                continue
            full_score, partial_score, read_start, read_end = \
                align_adapter(read_seq_end, adapter.end_sequence[1], *scoring_scheme_vals)
            if partial_score > end_threshold and read_start != 0 and \
                    read_end - read_start >= min_trim_size:
                trim_amount = (end_size - read_start) + extra_trim_size
                self.end_trim_amount = max(self.end_trim_amount, trim_amount)
                self.end_adapter_alignments.append((adapter, full_score, partial_score,
                                                    read_start, read_end))
            if check_barcodes and adapter.is_barcode() and \
                    adapter.barcode_direction() == forward_or_reverse:
                self.end_barcode_scores[adapter.get_barcode_name()] = full_score

    def find_middle_adapters(self, adapters, middle_threshold, extra_middle_trim_good_side,
                             extra_middle_trim_bad_side, scoring_scheme_vals,
                             start_sequence_names, end_sequence_names):
        """
        Aligns an adapter sequence to the whole read to find places where the read should be split.
        """
        masked_seq = self.get_seq_with_start_end_adapters_trimmed()
        valid_base = len(masked_seq)
        for adapter_name, adapter_seq in adapters:

            # We keep aligning adapters as long we get strong hits, so we can find multiple
            # occurrences in a single read.
            while True:
                # Check if all bases are masked
                if valid_base == 0:
                    break

                full_score, _, read_start, read_end = align_adapter(
                    masked_seq, adapter_seq, *scoring_scheme_vals)
                if full_score >= middle_threshold:
                    masked_seq = masked_seq[:read_start] + '-' * (read_end - read_start) + \
                                 masked_seq[read_end:]
                    valid_base -= read_end - read_start
                    self.middle_adapter_positions.update(range(read_start, read_end))

                    self.middle_hit_str += '  ' + adapter_name + ' (read coords: ' + \
                                           str(read_start) + '-' + str(read_end) + ', ' + \
                                           'identity: ' + '%.1f' % full_score + '%)\n'
                    self.middle_hit_id.append((adapter_name, full_score))
                    trim_start = read_start - extra_middle_trim_good_side
                    if adapter_name in start_sequence_names:
                        trim_start = read_start - extra_middle_trim_bad_side

                    trim_end = read_end + extra_middle_trim_good_side
                    if adapter_name in end_sequence_names:
                        trim_end = read_end + extra_middle_trim_bad_side

                    self.middle_trim_positions.update(range(trim_start, trim_end))
                else:
                    break

    def formatted_start_seq(self, end_size, extra_trim_size):
        """
        Returns the start of the read sequence, with any found adapters highlighted in red.
        """
        start_seq = self.seq[:end_size]
        if not self.start_trim_amount:
            return start_seq
        red_bases = self.start_trim_amount - extra_trim_size
        formatted_str = ''
        if red_bases:
            formatted_str = red(start_seq[:red_bases])
        formatted_str += yellow(start_seq[red_bases:red_bases+extra_trim_size])
        formatted_str += start_seq[red_bases+extra_trim_size:]
        return formatted_str

    def formatted_end_seq(self, end_size, extra_trim_size):
        """
        Returns the end of the read sequence, with any found adapters highlighted in red.
        """
        end_seq = self.seq[-end_size:]
        if not self.end_trim_amount:
            return end_seq
        red_bases = self.end_trim_amount - extra_trim_size
        formatted_str = ''
        if red_bases:
            formatted_str = red(end_seq[-red_bases:])
        formatted_str = yellow(end_seq[-(red_bases+extra_trim_size):-red_bases]) + formatted_str
        formatted_str = end_seq[:-(red_bases+extra_trim_size)] + formatted_str
        return formatted_str

    def formatted_whole_seq(self, extra_trim_size):
        """
        Returns the entire read sequence, with any found adapters highlighted in red.
        """
        if not self.start_trim_amount and not self.end_trim_amount:
            return self.seq

        red_start_bases, red_end_bases = 0, 0
        if self.start_trim_amount:
            red_start_bases = self.start_trim_amount - extra_trim_size
        if self.end_trim_amount:
            red_end_bases = self.end_trim_amount - extra_trim_size
        if red_start_bases + red_end_bases >= len(self.seq):
            return red(self.seq)

        formatted_start, formatted_end = '', ''
        if self.start_trim_amount:
            formatted_start = red(self.seq[:red_start_bases])
        if self.end_trim_amount:
            formatted_end = red(self.seq[-red_end_bases:])
        middle = self.seq[red_start_bases:len(self.seq)-red_end_bases]

        if len(middle) <= extra_trim_size * 2:
            middle = yellow(middle)
        else:
            if self.start_trim_amount:
                middle = yellow(middle[:extra_trim_size]) + middle[extra_trim_size:]
            if self.end_trim_amount:
                middle = middle[:-extra_trim_size] + yellow(middle[-extra_trim_size:])
        return formatted_start + middle + formatted_end

    def formatted_start_and_end_seq(self, end_size, extra_trim_size, check_barcodes):
        read_seq = ''
        if check_barcodes:
            start_name, start_id = self.best_start_barcode
            end_name, end_id = self.best_end_barcode
            read_seq += 'start: ' + start_name + ' (' + '%.1f' % start_id + '%), '
            read_seq += 'end: ' + end_name + ' (' + '%.1f' % end_id + '%), '
            read_seq += 'barcode call: ' + self.barcode_call + '   '
        if len(self.seq) <= 2 * end_size:
            read_seq += self.formatted_whole_seq(extra_trim_size)
        else:
            read_seq += (self.formatted_start_seq(end_size, extra_trim_size) + '...' +
                         self.formatted_end_seq(end_size, extra_trim_size))
        return read_seq

    def full_start_end_output(self, end_size, extra_trim_size, check_barcodes):
        def get_alignment_string(aln):
            return aln[0].name + ', full score=' + str(aln[1]) + ', partial score=' + \
                   str(aln[2]) + ', read position: ' + str(aln[3]) + '-' + str(aln[4])
        output = self.get_name() + '\n'
        output += '  start: ' + self.formatted_start_seq(end_size, extra_trim_size) + '...\n'
        if self.start_adapter_alignments:
            output += '    start alignments:\n'
            for a in self.start_adapter_alignments:
                output += '      ' + get_alignment_string(a) + '\n'
        output += '  end:   ...' + self.formatted_end_seq(end_size, extra_trim_size) + '\n'
        if self.end_adapter_alignments:
            output += '    end alignments:\n'
            for a in self.end_adapter_alignments:
                output += '      ' + get_alignment_string(a) + '\n'
        if check_barcodes:
            start_name, start_id = self.best_start_barcode
            end_name, end_id = self.best_end_barcode
            output += '  Barcodes:\n'
            all_start_barcodes_str = ', '.join([b[0] + ' (' + '%.1f' % b[1] + '%)'
                                                for b in self.start_barcode_scores.items()])
            all_end_barcodes_str = ', '.join([b[0] + ' (' + '%.1f' % b[1] + '%)'
                                              for b in self.end_barcode_scores.items()])
            output += '    start barcodes:        ' + all_start_barcodes_str + '\n'
            output += '    end barcodes:          ' + all_end_barcodes_str + '\n'
            output += '    best start barcode:    ' + start_name + ' (' + '%.1f' % start_id + '%)\n'
            output += '    best end barcode:      ' + end_name + ' (' + '%.1f' % end_id + '%)\n'
            if self.albacore_barcode_call is not None:
                output += '    albacore barcode call: ' + self.albacore_barcode_call + '\n'
            output += '    final barcode call:    ' + self.barcode_call + '\n'
        return output

    def formatted_middle_seq(self):
        """
        If a middle adapter was found, this returns the relevant part of the read sequence, with
        the adapter highlighted in red.
        """
        if not self.middle_adapter_positions:
            return

        trimmed_seq = self.get_seq_with_start_end_adapters_trimmed()

        range_start = max(0, min(self.middle_trim_positions) - 100)
        range_end = min(len(trimmed_seq),
                        max(self.middle_trim_positions) + 100)
        formatted_str = '' if range_start == 0 else '(' + str(range_start) + ' bp)...'

        last_colour = None
        for i in range(range_start, range_end):
            char_colour = None
            if i in self.middle_trim_positions:
                char_colour = 'yellow'
            if i in self.middle_adapter_positions:
                char_colour = 'red'
            if char_colour != last_colour:
                formatted_str += END_FORMATTING
                if char_colour == 'yellow':
                    formatted_str += YELLOW
                if char_colour == 'red':
                    formatted_str += RED

            formatted_str += trimmed_seq[i]
            last_colour = char_colour
        if last_colour is not None:
            formatted_str += END_FORMATTING

        formatted_str += '' if range_end == len(trimmed_seq) \
            else '...(' + str(len(trimmed_seq) - range_end) + ' bp)'
        return formatted_str

    def middle_adapter_results(self, verbosity):
        if not self.middle_adapter_positions:
            return ''
        results = self.get_name() + '\n' + self.middle_hit_str
        if verbosity > 1:
            results += self.formatted_middle_seq() + '\n'
        return results

    def determine_barcode(self, barcode_threshold, barcode_diff, require_two_barcodes):
        """
        This function works through the logic of choosing a barcode for the read based on the
        settings and the read's barcode alignments. It stores its result in self.barcode_call.
        """
        start_barcode_scores = sorted(self.start_barcode_scores.items(), reverse=True,
                                      key=lambda x: x[1])
        end_barcode_scores = sorted(self.end_barcode_scores.items(), reverse=True,
                                    key=lambda x: x[1])

        if len(start_barcode_scores) >= 1:
            self.best_start_barcode = start_barcode_scores[0]
        if len(start_barcode_scores) >= 2:
            self.second_best_start_barcode = start_barcode_scores[1]
        if len(end_barcode_scores) >= 1:
            self.best_end_barcode = end_barcode_scores[0]
        if len(end_barcode_scores) >= 2:
            self.second_best_end_barcode = end_barcode_scores[1]

        try:
            # If the user set --require_two_barcodes, then the criteria are much more stringent.
            # Both the start and end barcodes need to be over the threshold, they both need to be
            # sufficiently better than their second-best barcode hit, and they need to match.
            if require_two_barcodes:
                start_over_threshold = (self.best_start_barcode[1] >= barcode_threshold)
                end_over_threshold = (self.best_end_barcode[1] >= barcode_threshold)
                start_good_diff = (self.best_start_barcode[1] >=
                                   self.second_best_start_barcode[1] + barcode_diff)
                end_good_diff = (self.best_end_barcode[1] >=
                                 self.second_best_end_barcode[1] + barcode_diff)
                start_end_match = (self.best_start_barcode[0] == self.best_end_barcode[0])

                assert (start_over_threshold and end_over_threshold and
                        start_good_diff and end_good_diff and start_end_match)
                self.barcode_call = self.best_start_barcode[0]

            # If the user didn't set --require_two_barcodes, then the criteria aren't so strict.
            # The start/end barcodes are analysed all together.
            else:
                # Combine the start and end barcodes into a single list (i.e. we no longer care
                # whether the hit was at the start or end of the read), only keeping the best score
                # for each barcode.
                all_barcode_scores = []
                included_barcodes = set()
                for name, score in sorted(start_barcode_scores + end_barcode_scores, reverse=True,
                                          key=lambda x: x[1]):
                    if name not in included_barcodes:
                        all_barcode_scores.append((name, score))
                        included_barcodes.add(name)

                if len(all_barcode_scores) >= 1:
                    best_overall_barcode = all_barcode_scores[0]
                else:
                    best_overall_barcode = ('none', 0.0)
                if len(all_barcode_scores) >= 2:
                    second_best_overall_barcode = all_barcode_scores[1]
                else:
                    second_best_overall_barcode = ('none', 0.0)

                over_threshold = (best_overall_barcode[1] >= barcode_threshold)
                good_diff = (best_overall_barcode[1] >=
                             second_best_overall_barcode[1] + barcode_diff)
                assert over_threshold
                assert good_diff

                self.barcode_call = best_overall_barcode[0]

        except AssertionError:
            self.barcode_call = 'none'

        # If the read has been binned by Albacore, then Porechop and Albacore must agree on the
        # barcode. If they don't, the read is unclassified.
        if self.albacore_barcode_call is not None and \
                self.barcode_call != self.albacore_barcode_call:
            self.barcode_call = 'none'


def align_adapter(read, adapter, match=3, mismatch=-6, gap_open=5, gap_extend=2):
    """
    SSW local alignment of (read, adapter)
    """
    score_size = 1
    matrix = parasail.matrix_create("ACGT", match, mismatch)
    profile = parasail.ssw_init(read, matrix, score_size)
    hit = parasail.ssw_profile(profile, adapter, gap_open, gap_extend)

    if hit is None or hit.read_end1 == hit.read_begin1: # If no hit
        read_start = -1
        read_end = 0
        aligned_region_percent_identity = 0.0
        full_adapter_percent_identity = 0.0
    else:
        read_start = hit.read_begin1
        read_end = hit.read_end1+1

        _read = read[hit.read_begin1:hit.read_end1+1]
        _adapter = adapter[hit.ref_begin1:hit.ref_end1+1]

        region_distance = edlib.align(_read, _adapter)['editDistance']
        aligned_region_percent_identity = 100 * (1 - region_distance / len(_adapter))

        # adapter_distance = edlib.align(_read, adapter)['editDistance']
        # full_adapter_percent_identity = 100 * (1 - adapter_distance / len(adapter))

        # Re-alignment of candidate adapter
        _read = read[hit.read_begin1-hit.ref_begin1:hit.read_end1+len(adapter)-hit.ref_end1]
        adapter_distance = edlib.align(_read, adapter)['editDistance']
        full_adapter_percent_identity = 100 * (1 - adapter_distance / len(adapter))

    return full_adapter_percent_identity, aligned_region_percent_identity, read_start, read_end


def align_adapter_global(read, adapter, match=3, mismatch=-6, gap_open=5, gap_extend=2):
    """
    Global alignment of (read, adapter)
    """
    matrix = parasail.matrix_create("ACGT", match, mismatch)
    hit = parasail.sg_qx_trace(read, adapter, gap_open, gap_extend, matrix)
    # hit = parasail.sg_trace(read, adapter, gap_open, gap_extend, matrix)
    if hit is None: # If no hit
        read_start = -1
        read_end = 0
        aligned_region_percent_identity = 0.0
        full_adapter_percent_identity = 0.0
    else:
        adapter_str = hit.traceback.ref
        read_start = len(adapter_str) - len(adapter_str.lstrip("-"))
        read_end = len(adapter_str) - len(adapter_str.rstrip("-"))

        # _read = read[hit.read_begin1:hit.read_end1+1]
        # _adapter = adapter[hit.ref_begin1:hit.ref_end1+1]

        n_match = sum([i == k for i, j, k in zip(
            hit.traceback.query, hit.traceback.comp, hit.traceback.ref)])
        aligned_region_percent_identity = 100 * n_match / len(adapter_str.strip("-"))
        full_adapter_percent_identity = 100 * n_match / len(adapter_str.strip("-"))

    return full_adapter_percent_identity, aligned_region_percent_identity, read_start, read_end


def align_adapter_set(read, adapter_set, end_size, scoring_scheme_vals):
    """
    This function aligns the adapter to the reads and updates the best score for the adapter.
    This is not to determine where to trim the reads, but rather to figure out which adapter
    sets are present in the data.
    """
    read_seq_start = read.seq[:end_size]
    score, _, _, _ = align_adapter(read_seq_start, adapter_set.start_sequence[1],
                                      *scoring_scheme_vals)
    adapter_set.best_start_score = max(adapter_set.best_start_score, score)
    if adapter_set.end_sequence:
        read_seq_end = read.seq[-end_size:]
        score, _, _, _ = align_adapter(read_seq_end, adapter_set.end_sequence[1],
                                          *scoring_scheme_vals)
        adapter_set.best_end_score = max(adapter_set.best_end_score, score)


def find_matching_adapter_sets(check_reads, verbosity, end_size, scoring_scheme_vals,
                               adapter_threshold, threads, search_adapters=None):
    """
    Aligns all of the adapter sets to the start/end of reads to see which (if any) matches best.
    """
    read_count = len(check_reads)
    logger.info('Looking for known adapter sets')

    search_adapter_count = len(search_adapters)

    # multi-thread is not useful here due to the GIL
    for read in tqdm(check_reads, ncols=75):
        for adapter_set in search_adapters:
            align_adapter_set(read, adapter_set, end_size, scoring_scheme_vals)
    sys.stderr.write('\n')

    return [x for x in search_adapters if x.best_start_or_end_score() >= adapter_threshold]


def exclude_end_adapters_for_rapid(matching_sets):
    """
    Rapid reads shouldn't have end adapters, so we don't want to look for them if this seems to be
    a rapid read set.
    """
    if 'Rapid adapter' in [x.name for x in matching_sets]:
        for s in matching_sets:
            s.end_sequence = None
    return matching_sets


def fix_up_1d2_sets(matching_sets):
    """
    The 1D^2 adapters share a lot of common sequence with the old SQK-MAP006_short adapters, so if
    there are strong 1D^2 hits, we can exclude the redundant SQK-MAP006_short adapters.
    """
    matching_set_names = [x.name for x in matching_sets]
    if '1D^2 part 1' in matching_set_names and '1D^2 part 2' in matching_set_names and \
            'SQK-MAP006 Short' in matching_set_names:
        part_1_score = [x for x in matching_sets
                        if x.name == '1D^2 part 1'][0].best_start_or_end_score()
        part_2_score = [x for x in matching_sets
                        if x.name == '1D^2 part 2'][0].best_start_or_end_score()
        sqk_score = [x for x in matching_sets
                     if x.name == 'SQK-MAP006 Short'][0].best_start_or_end_score()
        if part_1_score >= sqk_score and part_2_score >= sqk_score:
            matching_sets = [
                x for x in matching_sets if x.name != 'SQK-MAP006 Short']
    return matching_sets


def add_full_barcode_adapter_sets(matching_sets):
    """
    This function adds some new 'full' adapter sequences based on what was already found. For
    example, if the ligation adapters and the reverse barcode adapters are found, it assumes we are
    looking at a native barcoding run and so it adds the complete native barcoding adapter
    sequences (with the barcode's upstream and downstream context included).
    """
    matching_set_names = [x.name for x in matching_sets]

    for i in range(1, 97):

        # Native barcode full sequences
        if all(x in matching_set_names
               for x in ['SQK-NSK007', 'Barcode ' + str(i) + ' (reverse)']):
            matching_sets.append(make_full_native_barcode_adapter(i))

        # Rapid barcode full sequences
        if all(x in matching_set_names
               for x in ['SQK-NSK007', 'Rapid', 'Barcode ' + str(i) + ' (forward)']):
            matching_sets.append(make_full_rapid_barcode_adapter(i))

    return matching_sets


def display_adapter_set_results(search_adapters, matching_sets):
    from porechopx.misc import print_table
    table = [['Set', 'Best read start %ID', 'Best read end %ID']]
    row_colours = {}
    matching_set_names = [x.name for x in matching_sets]
    for adapter_set in search_adapters:
        start_score = '%.1f' % adapter_set.best_start_score
        end_score = '%.1f' % adapter_set.best_end_score
        table.append([adapter_set.name, start_score, end_score])
        if adapter_set.name in matching_set_names:
            row_colours[len(table) - 1] = 'green'
    print_table(table, sys.stderr, alignments='LRR', row_colour=row_colours,
                fixed_col_widths=[35, 8, 8])
    sys.stderr.write('\n')


def choose_barcoding_kit(adapter_sets):
    """
    If the user is sorting reads by barcode bin, choose one barcode configuration (rev comp
    barcodes at the start of the read or at the end of the read) and ignore the other.
    """
    forward_barcodes = 0
    reverse_barcodes = 0
    for adapter_set in adapter_sets:
        score = adapter_set.best_start_or_end_score()
        if 'Native' in adapter_set.name:
            reverse_barcodes += score
        elif 'Barcode' in adapter_set.name and '(forward)' in adapter_set.name:
            forward_barcodes += score
        elif 'Barcode' in adapter_set.name and '(reverse)' in adapter_set.name:
            reverse_barcodes += score

    if forward_barcodes > reverse_barcodes:
        logger.info('Barcodes determined to be in forward orientation')
        return 'forward'
    elif reverse_barcodes > forward_barcodes:
        logger.info('Barcodes determined to be in reverse orientation',)
        return 'reverse'
    else:
        return None


def add_number_to_read_name(read_name, number):
    if ' ' not in read_name:
        return read_name + '_' + str(number)
    else:
        return read_name.replace(' ', '_' + str(number) + ' ', 1)
